"""
Unit tests for `toddler.plugins.cleaning_packager_groups`
"""

import logging
from unittest.mock import MagicMock, patch

from journal_to_fedora_messaging_messages.ipa import IpaGroupRemoveMemberV1
import pytest

from toddlers.exceptions import PagureError
import toddlers.plugins.cleaning_packager_groups as cleaning_packager_groups


@pytest.fixture
def group_remove_member_message():
    return IpaGroupRemoveMemberV1(
        {
            "IPA_API_PARAMS": (
                '{"cn": "packager", "all": false, "raw": false, "version": "2.254", '
                '"no_members": false, "user": ["lenkaseg"]}'
            ),
            "__REALTIME_TIMESTAMP": "1742838354523018",
            "IPA_API_ACTOR": "agent@EXAMPLE.COM",
            "IPA_API_COMMAND": "group_remove_member",
            "IPA_API_RESULT": "SUCCESS",
            "MESSAGE": (
                "[IPA.API] agent@EXAMPLE.COM: group_remove_member: SUCCESS [ldap2_139734790206096] "
                '{"cn": "packager", "all": false, "raw": false, "version": "2.254", '
                '"no_members": false, "user": ["lenkaseg"]}'
            ),
            "_HOSTNAME": "ipa.example.com",
            "_COMM": "httpd",
            "PRIORITY": "5",
        }
    )


class TestAcceptTopic:
    """
    Test class for `toddler.plugins.cleaning_packager_groups.CleanPackagerGroups.accepts_topic`
    method.
    """

    toddler_cls = cleaning_packager_groups.CleanPackagerGroups

    def test_accepts_topic_invalid(self, toddler):
        """
        Assert that invalid topic is not accepted.
        """
        assert toddler.accepts_topic("foo.bar") is False

    @pytest.mark.parametrize(
        "topic",
        [
            "org.fedoraproject.*.ipa.group_remove_member.v1",
            "org.fedoraproject.prod.ipa.group_remove_member.v1",
            "org.fedoraproject.stg.ipa.group_remove_member.v1",
        ],
    )
    def test_accepts_topic_valid(self, topic, toddler):
        """
        Assert that valid topics are accepted.
        """
        assert toddler.accepts_topic(topic)


class TestProcess:
    """
    Test class for `toddler.plugins.cleaning_packager_groups.CleanPackagerGroups.process` method
    """

    def setup_method(self):
        """Initialize toddler."""
        self.toddler_cls = cleaning_packager_groups.CleanPackagerGroups()
        self.toddler_cls._ipa_session = MagicMock()

    @patch("toddlers.utils.pagure.set_pagure")
    def test_not_packager_group(
        self, mock_set_pagure, group_remove_member_message, caplog
    ):
        """
        Assert that if the group in the message is not "packager", plugin stops.
        """
        mock_set_pagure.return_value.remove_member_from_group.return_value = None
        mock_set_pagure.return_value.get_all_groups.return_value = ["group1", "group2"]
        caplog.set_level(logging.INFO)
        config = {
            "pagure_url": "https://pagure.io",
            "pagure_stg": "https://stg.pagure.io",
            "watched_groups": ["packager"],
        }
        group_remove_member_message.body["IPA_API_PARAMS"] = (
            '{"cn": "not-packager", "all": false, "raw": false, "version": "2.254", '
            '"no_members": false, "user": ["lenkaseg"]}'
        )
        self.toddler_cls.process(config, group_remove_member_message)
        assert (
            caplog.records[-1].message
            == "User(s) lenkaseg was not removed from packager group, but "
            "from not-packager group, nothing to do."
        )

    @pytest.mark.parametrize(
        "distgit_groups,ipa_groups,messages",
        (
            (
                ["group1", "group2"],
                {"result": {"memberof_group": ["group1", "group3"]}},
                (
                    "Fetching all distgit groups:",
                    "Distgit groups found: ['group1', 'group2']",
                    "User lenkaseg was removed from packager group, removing from "
                    "packager-related groups as well.",
                    "User lenkaseg removed from distgit group: packager",
                    "Fetching groups user lenkaseg is member of in ipa:",
                    "Ipa groups: ['group1', 'group3']",
                    "User lenkaseg should be removed from following groups: ['group1']",
                    "User lenkaseg removed from ipa group: group1",
                    "User lenkaseg removed from distgit group: group1",
                ),
            ),
            (
                None,
                {"result": {"memberof_group": ["group1", "group3"]}},
                (
                    "Fetching all distgit groups:",
                    "Distgit groups found: None",
                    "No distgit groups found, bailing.",
                ),
            ),
            (
                ["group1", "group2"],
                {"result": {"memberof_group": ["group3", "group4"]}},
                (
                    "Fetching all distgit groups:",
                    "Distgit groups found: ['group1', 'group2']",
                    "User lenkaseg was removed from packager group, removing from "
                    "packager-related groups as well.",
                    "User lenkaseg removed from distgit group: packager",
                    "Fetching groups user lenkaseg is member of in ipa:",
                    "Ipa groups: ['group3', 'group4']",
                    "Seems that user lenkaseg is not member of any groups shared "
                    "between ipa and distgit",
                ),
            ),
            (
                ["group1", "group2"],
                {"result": {"memberof_group": []}},
                (
                    "Fetching all distgit groups:",
                    "Distgit groups found: ['group1', 'group2']",
                    "User lenkaseg was removed from packager group, removing from "
                    "packager-related groups as well.",
                    "User lenkaseg removed from distgit group: packager",
                    "Fetching groups user lenkaseg is member of in ipa:",
                    "Ipa groups: []",
                    "Seems that user lenkaseg is not member of any groups shared "
                    "between ipa and distgit",
                ),
            ),
        ),
        ids=(
            "distgit_groups_found",
            "distgit_group_not_found",
            "no intersecting distgit/ipa groups",
            "ipa_groups_not_found",
        ),
    )
    @patch("toddlers.utils.pagure.set_pagure")
    def test_is_packager_group_and_post_is_ok(
        self,
        mock_pagure,
        distgit_groups,
        ipa_groups,
        messages,
        group_remove_member_message,
        caplog,
    ):
        """
        Assert if group is packager, the plugin removes the user from related distgit groups.
        """
        mock_pagure_obj = MagicMock()
        mock_pagure_obj.get_all_groups.return_value = distgit_groups
        mock_pagure_obj.remove_member_from_group.return_value = None
        mock_pagure.return_value = mock_pagure_obj

        self.toddler_cls._ipa_session.Command.user_show.return_value = ipa_groups
        caplog.set_level(logging.INFO)
        config = {
            "pagure_url": "https://example.io",
            "pagure_stg": "https://example.stg.io",
            "watched_groups": ["packager"],
        }
        self.toddler_cls.process(config, group_remove_member_message)
        recorded_messages = tuple(r.message for r in caplog.records)
        print(recorded_messages)
        assert recorded_messages == messages

    @pytest.mark.parametrize(
        "ipa_remove_member,error,message_1,message_2",
        (
            (
                Exception,
                None,
                "User lenkaseg removed from distgit group: group1",
                "Error while removing user lenkaseg from ipa group group1",
            ),
            (
                None,
                PagureError,
                "Error while removing user lenkaseg from distgit group group1",
                "User lenkaseg removed from ipa group: group1",
            ),
        ),
        ids=("ipa_error", "distgit_error"),
    )
    @patch("toddlers.plugins.cleaning_packager_groups.pagure.set_pagure")
    def test_is_packager_group_and_post_is_not_ok(
        self,
        mock_set_pagure,
        ipa_remove_member,
        error,
        message_1,
        message_2,
        group_remove_member_message,
        caplog,
    ):
        """
        Assert correct behavior when encountered with pagure error.
        """
        mock_set_pagure.return_value.get_all_groups.return_value = ["group1", "group2"]
        mock_set_pagure.return_value.remove_member_from_group.side_effect = error

        self.toddler_cls._ipa_session.Command.user_show.return_value = {
            "result": {"memberof_group": ["group1", "group3"]}
        }

        self.toddler_cls._ipa_session.Command.group_remove_member.side_effect = (
            ipa_remove_member
        )

        caplog.set_level(logging.INFO)
        config = {
            "pagure_url": "https://example.io",
            "pagure_stg": "https://example.stg.io",
            "pagure_api_key": "pagure_key",
            "watched_groups": ["packager"],
        }

        self.toddler_cls.process(config, group_remove_member_message)
        self.toddler_cls._ipa_session.Command.user_show.assert_called_with(
            uid="lenkaseg"
        )
        assert caplog.records[-1].message == message_1
        assert caplog.records[-2].message == message_2
