"""
Unit tests for `toddler.plugins.unretire_packages`
"""

import logging
from unittest.mock import MagicMock, Mock, patch

import arrow
import git
import koji
from pagure_messages.issue_schema import IssueNewV1
import pytest

from toddlers.exceptions import ValidationError
import toddlers.plugins.unretire_packages as unretire_packages


class TestAcceptTopic:
    """
    Test class for `toddler.plugins.unretire_packages.UnretirePackages.accepts_topic`
    method.
    """

    toddler_cls = unretire_packages.UnretirePackages

    def test_accepts_topic_invalid(self, toddler):
        """
        Assert that invalid topic is not accepted.
        """
        assert toddler.accepts_topic("foo.bar") is False

    @pytest.mark.parametrize(
        "topic",
        [
            "io.pagure.*.pagure.issue.new",
            "io.pagure.stg.pagure.issue.new",
            "io.pagure.prod.pagure.issue.new",
        ],
    )
    def test_accepts_topic_valid(self, topic, toddler):
        assert toddler.accepts_topic(topic)


class TestProcess:
    """
    Test class for `toddlers.plugins.unretire_packages.UnretirePackages.process` method
    """

    toddler_cls = unretire_packages.UnretirePackages

    def test_process_invalid_project(self, caplog, toddler):
        """
        Asserts that messages from other projects than `` will be skipped.
        """
        caplog.set_level(logging.INFO)

        body = {
            "project": {"fullname": "foo/bar"},
            "issue": {"user": {"name": "zlopez"}},
        }
        msg = IssueNewV1(body=body)

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages.process_ticket"
        ) as mock_process_ticket:
            toddler.process({}, msg)
            mock_process_ticket.assert_not_called()

        assert (
            caplog.records[-1].message
            == "The message doesn't belong to project releng/fedora-scm-requests. Skipping message."
        )

    def test_process_issue_not_open(self, caplog, toddler):
        """
        Assert that message with closed issues will be skipped.
        """
        caplog.set_level(logging.INFO)

        body = {
            "project": {"fullname": unretire_packages.PROJECT_NAMESPACE},
            "issue": {
                "id": 100,
                "status": "Closed",
            },
        }
        msg = IssueNewV1(body=body)

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages.process_ticket"
        ) as mock_process_ticket:
            toddler.process({}, msg)

            mock_process_ticket.assert_not_called()

        assert (
            caplog.records[-1].message == "The issue 100 is not open. Skipping message."
        )

    def test_process_keyword_not_in_title(self, caplog, toddler):
        """
        Assert that message without keyword in title will be skipped.
        """
        caplog.set_level(logging.INFO)

        body = {
            "project": {"fullname": unretire_packages.PROJECT_NAMESPACE},
            "issue": {
                "id": 100,
                "status": "Open",
                "title": "New Repo",
            },
        }
        msg = IssueNewV1(body=body)

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages.process_ticket"
        ) as mock_process_ticket:
            toddler.process({}, msg)

            mock_process_ticket.assert_not_called()

        assert (
            caplog.records[-1].message
            == "The issue doesn't contain keyword 'unretire' in the title 'New Repo'"
        )

    @patch("toddlers.utils.requests.make_session")
    @patch("toddlers.utils.bodhi.set_bodhi")
    @patch("toddlers.utils.pagure.set_pagure")
    @patch("toddlers.utils.bugzilla_system.set_bz")
    @patch("koji.ClientSession")
    def test_process_base_exception_appeared(
        self, mock_koji, mock_bz, mock_pagure, mock_bodhi, session_mock, caplog, toddler
    ):
        """
        Assert that if base exception appeared, it will be processed correctly.
        """
        issue = {
            "id": 100,
            "status": "Open",
            "title": "Unretire package_name",
            "user": {"name": "amedvede"},
        }
        body = {
            "project": {"fullname": unretire_packages.PROJECT_NAMESPACE},
            "issue": issue,
        }
        config = {
            "dist_git_url": "https://src.fedoraproject.org",
            "dist_git_token": "Private API Key",
            "temp_folder": "/var/tmp",
        }
        msg = IssueNewV1(body=body)

        mock_pagure_io = Mock()
        mock_pagure.return_value = mock_pagure_io

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages.process_ticket"
        ) as mock_process_ticket:
            mock_process_ticket.side_effect = Exception("test")
            toddler.process(config, msg)
            mock_process_ticket.assert_called_with(issue)

        mock_pagure_io.add_comment_to_issue.assert_called_once()

    @patch("toddlers.utils.requests.make_session")
    @patch("toddlers.utils.bodhi.set_bodhi")
    @patch("toddlers.utils.pagure.set_pagure")
    @patch("toddlers.utils.bugzilla_system.set_bz")
    @patch("koji.ClientSession")
    def test_process(
        self, mock_koji, mock_bugzilla, mock_pagure, mock_bodhi, session_mock, toddler
    ):
        """
        Assert that message toddler will be initialized correctly, if message passes
        initial processing.
        """
        issue = {
            "id": 100,
            "status": "Open",
            "title": "Unretire package",
            "user": {"name": "amedvede"},
        }
        body = {
            "project": {"fullname": unretire_packages.PROJECT_NAMESPACE},
            "issue": issue,
        }
        msg = IssueNewV1(body=body)
        config = {
            "dist_git_url": "dist_git_url",
            "dist_git_token": "dist_git_token",
            "kojihub_url": "koji",
            "pagure_url": "https://pagure.io",
            "pagure_api_key": "some api key",
            "bugzilla_url": "https://bugzilla.redhat.com/",
            "bugzilla_api_key": "some api key",
        }

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages.process_ticket"
        ) as mock_process_ticket:
            toddler.process(config, msg)

            mock_process_ticket.assert_called_with(issue)

        mock_pagure.assert_called_with(config)
        mock_bugzilla.assert_called_with(config)
        mock_bodhi.assert_called_with(config)
        session_mock.assert_called()


class TestProcessTicket:
    """
    Test class for `toddlers.plugin.unretire_packages.UnretirePackages.process_ticket`
    method.
    """

    def setup_method(self):
        """
        Initialize toddler
        """
        self.toddler = unretire_packages.UnretirePackages()
        self.toddler.temp_dir = Mock()
        self.toddler.git_repo = MagicMock()
        self.toddler.pagure_io = MagicMock()
        self.toddler.pagure_user = "bot"
        self.toddler.koji_session = MagicMock()
        self.toddler.bodhi = MagicMock()
        self.toddler.oidc_distgit_token = "token"

    @patch("json.loads")
    def test_process_ticket_issue_body_is_invalid_json(self, json_loads):
        json_loads.side_effect = ValueError("test")
        issue = {"id": 100, "content": "issue body", "full_url": "ticket_url"}
        err_msg = "Invalid JSON provided"

        self.toddler.process_ticket(issue)
        self.toddler.pagure_io.close_issue.assert_called_once_with(
            issue["id"],
            namespace=unretire_packages.PROJECT_NAMESPACE,
            message=err_msg,
            reason="Invalid",
        )

    @patch("json.loads")
    def test_process_ticket_package_doesnt_exist(self, json_loads_mock):
        issue_body = {
            "name": "pkg_name",
            "type": "rpms",
            "maintainer": "amedvede",
        }
        json_loads_mock.return_value = issue_body
        issue = {
            "id": 100,
            "content": "issue body",
            "full_url": "ticket_url",
        }
        err_msg = "Package repository doesnt exist. Try to repeat request."
        self.toddler._does_url_exist = MagicMock(return_value=False)

        self.toddler.process_ticket(issue)
        self.toddler.pagure_io.close_issue.assert_called_once_with(
            issue["id"],
            namespace=unretire_packages.PROJECT_NAMESPACE,
            message=err_msg,
            reason="Invalid",
        )

    @patch("json.loads")
    @patch("toddlers.plugins.unretire_packages.git.clone_repo")
    @patch("tempfile.TemporaryDirectory")
    @patch(
        "toddlers.plugins.unretire_packages.UnretirePackages._does_url_exist",
        return_value=True,
    )
    def test_process_ticket_git_error_happened(
        self,
        mock_does_url_exist,
        mock_tmp_dir,
        mock_clone_repo,
        json_loads_mock,
        caplog,
    ):
        caplog.set_level(logging.INFO)
        issue_body = {
            "name": "pkg_name",
            "type": "rpms",
            "maintainer": "amedvede",
        }
        json_loads_mock.return_value = issue_body
        issue = {
            "id": 100,
            "content": "issue body",
            "full_url": "ticket_url",
        }
        err_msg = "Something went wrong during cloning git repository."
        mock_clone_repo.side_effect = git.GitCommandError("Error cloning")

        self.toddler.process_ticket(issue)
        assert caplog.records[-1].message == err_msg
        self.toddler.pagure_io.close_issue.assert_called_once_with(
            issue["id"],
            namespace=unretire_packages.PROJECT_NAMESPACE,
            message=err_msg,
            reason="Invalid",
        )

    @patch("json.loads")
    @patch("toddlers.plugins.unretire_packages.git.clone_repo")
    @patch("tempfile.TemporaryDirectory")
    @patch(
        "toddlers.plugins.unretire_packages.UnretirePackages._does_url_exist",
        return_value=True,
    )
    def test_process_ticket_package_is_not_ready_for_unretirement(
        self,
        mock_does_url_exist,
        mock_tmp_dir,
        mock_clone_repo,
        json_loads_mock,
        caplog,
    ):
        caplog.set_level(logging.DEBUG)
        issue_body = {
            "name": "pkg_name",
            "type": "rpms",
            "maintainer": "amedvede",
            "branches": ["f37", "f38", "f39", "f40"],
            "review_bugzilla": "242142144",
        }
        json_loads_mock.return_value = issue_body
        issue = {
            "id": 100,
            "content": "issue body",
            "full_url": "ticket_url",
        }
        last_log_msg = "Verifying if package is ready for unretirement."
        self.toddler.bodhi.get_active_branches.return_value = ["f38", "f39", "f40"]
        mock_clone_repo.return_value.does_branch_exist.return_value = True
        mock_clone_repo.return_value.does_branch_contains_file.return_value = True
        self.toddler._is_package_ready_for_unretirement = MagicMock(return_value=False)

        self.toddler.process_ticket(issue)
        assert caplog.records[-1].message == last_log_msg
        self.toddler._is_package_ready_for_unretirement.assert_called_with(
            issue_id=issue["id"],
            branches=["f38", "f39", "f40"],
            review_bugzilla="242142144",
        )
        self.toddler.git_repo.revert_last_commit.assert_not_called()

    @patch("toddlers.plugins.unretire_packages.git.clone_repo")
    @patch("tempfile.TemporaryDirectory")
    @patch("json.loads")
    def test_process_ticket_no_maintainer_fas(
        self, json_loads_mock, temp_dir_mock, mock_clone_repo, caplog
    ):
        caplog.set_level(logging.INFO)
        issue_body = {
            "name": "pkg_name",
            "type": "rpms",
            "maintainer": "",
            "branches": ["f37", "f38", "f39", "f40"],
            "review_bugzilla": "242142144",
        }
        json_loads_mock.return_value = issue_body
        issue = {
            "id": 100,
            "content": "issue body",
            "full_url": "ticket_url",
        }
        err_msg = "Package is ophaned, but maintainer fas is not provided."
        self.toddler._does_url_exist = MagicMock(return_value=True)
        self.toddler.bodhi.get_active_branches.return_value = ["f38", "f39", "f40"]
        mock_clone_repo.return_value.does_branch_exist.return_value = True
        mock_clone_repo.return_value.does_branch_contains_file.return_value = True
        self.toddler._is_package_ready_for_unretirement = MagicMock(return_value=True)
        self.toddler._check_tags_to_unblock = MagicMock(return_value=False)
        self.toddler.pagure_io.is_project_orphaned = MagicMock(return_value=True)

        self.toddler.process_ticket(issue)
        self.toddler.pagure_io.close_issue.assert_called_once_with(
            issue_id=issue["id"],
            namespace=unretire_packages.PROJECT_NAMESPACE,
            message=err_msg,
            reason="Invalid",
        )

    @patch("toddlers.plugins.unretire_packages.git.clone_repo")
    @patch("tempfile.TemporaryDirectory")
    @patch("json.loads")
    def test_process_ticket_koji_error(
        self, json_loads_mock, temp_dir_mock, mock_clone_repo, caplog
    ):
        caplog.set_level(logging.INFO)
        issue_body = {
            "name": "pkg_name",
            "type": "rpms",
            "maintainer": "sososo",
            "branches": ["f37", "f38", "f39", "f40"],
            "review_bugzilla": "242142144",
        }
        json_loads_mock.return_value = issue_body
        issue = {
            "id": 100,
            "content": "issue body",
            "full_url": "ticket_url",
        }
        err_msg = "Not able to unblock `f38` tag on koji."
        self.toddler._does_url_exist = MagicMock(return_value=True)
        self.toddler.bodhi.get_active_branches.return_value = ["f38", "f39", "f40"]
        mock_clone_repo.return_value.does_branch_exist.return_value = True
        mock_clone_repo.return_value.does_branch_contains_file.return_value = True
        self.toddler._is_package_ready_for_unretirement = MagicMock(return_value=True)
        self.toddler._check_tags_to_unblock = MagicMock(return_value=True)
        self.toddler.koji_session.packageListUnblock = MagicMock(
            side_effect=koji.GenericError
        )
        self.toddler.pagure_io.is_project_orphaned = MagicMock()

        self.toddler.process_ticket(issue)
        self.toddler.pagure_io.close_issue.assert_called_once_with(
            issue_id=issue["id"],
            namespace=unretire_packages.PROJECT_NAMESPACE,
            message=err_msg,
            reason="Invalid",
        )
        self.toddler.pagure_io.is_project_orphaned.assert_not_called()

    @patch("toddlers.plugins.unretire_packages.git.clone_repo")
    @patch("tempfile.TemporaryDirectory")
    @patch("json.loads")
    def test_process_ticket_success(
        self, json_loads_mock, temp_dir_mock, mock_clone_repo, caplog
    ):
        caplog.set_level(logging.INFO)
        issue_body = {
            "name": "pkg_name",
            "type": "rpms",
            "maintainer": "amedvede",
            "branches": ["f37", "f38", "f39", "f40"],
            "review_bugzilla": "242142144",
        }
        json_loads_mock.return_value = issue_body
        issue = {
            "id": 100,
            "content": "issue body",
            "full_url": "ticket_url",
        }
        self.toddler._does_url_exist = MagicMock(return_value=True)
        self.toddler.bodhi.get_active_branches.return_value = ["f38", "f39", "f40"]
        mock_clone_repo.return_value.does_branch_exist.return_value = True
        mock_clone_repo.return_value.does_branch_contains_file.return_value = True
        self.toddler._is_package_ready_for_unretirement = MagicMock(return_value=True)
        self.toddler._check_tags_to_unblock = MagicMock(return_value=True)
        self.toddler.koji_session.packageListUnblock = MagicMock()
        self.toddler._unblock_tags_on_koji = MagicMock()
        self.toddler.pagure_io.is_project_orphaned = MagicMock(return_value=True)
        self.toddler.pagure_io.assign_maintainer_to_project = MagicMock()

        self.toddler.process_ticket(issue)
        self.toddler._is_package_ready_for_unretirement.assert_called_once_with(
            issue_id=issue["id"],
            branches=["f38", "f39", "f40"],
            review_bugzilla="242142144",
        )
        self.toddler._check_tags_to_unblock.assert_called_once_with(
            ["f38", "f39", "f40"], "pkg_name"
        )
        self.toddler.pagure_io.is_project_orphaned.assert_called_once_with(
            namespace=issue_body["type"], repo=issue_body["name"]
        )
        self.toddler.pagure_io.assign_maintainer_to_project.assert_called_once_with(
            namespace=issue_body["type"],
            repo=issue_body["name"],
            maintainer_fas="amedvede",
        )
        package = "rpms/pkg_name"
        assert caplog.records[-1].message == "Package {0} is assigned to {1}".format(
            package, issue_body["maintainer"]
        )


class TestIsPackageReadyForUnretirement:
    """
    Test class for `toddlers.plugin.unretire_packages.UnreturePackages
    ._is_package_ready_for_unretirement` method.
    """

    def setup_method(self):
        """
        Initialize toddler
        """
        self.toddler = unretire_packages.UnretirePackages()
        self.toddler.pagure_io = Mock()

    @patch(
        "toddlers.plugins.unretire_packages.UnretirePackages." "_verify_bugzilla_ticket"
    )
    @patch(
        "toddlers.plugins.unretire_packages.UnretirePackages."
        "_verify_package_not_retired_for_reason"
    )
    def test_is_package_ready_for_unretirement_success(
        self, verify_reason_mock, need_to_check_bz_mock
    ):
        """
        Assert that method process correctly.
        """
        issue_id = 55
        branches = ["f32", "f35"]
        review_bugzilla = "242142144"

        assert (
            self.toddler._is_package_ready_for_unretirement(
                issue_id, branches, review_bugzilla
            )
            is True
        )
        verify_reason_mock.assert_called_with(branches=branches)
        need_to_check_bz_mock.assert_called_with(
            review_bugzilla=review_bugzilla, branches=branches
        )

    def test_is_package_ready_for_unretirement_legal_issue(self):
        """
        Assert that package won't be verified if it has legal of license issue
        """
        issue_id = 55
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "242142144"
        err_msg = "Package was retired for a reason: legal of license issue."

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages"
            "._verify_package_not_retired_for_reason",
            side_effect=ValidationError(err_msg),
        ):
            assert (
                self.toddler._is_package_ready_for_unretirement(
                    issue_id, branches, review_bugzilla
                )
                is False
            )
            self.toddler.pagure_io.close_issue.assert_called_with(
                issue_id=issue_id,
                namespace=unretire_packages.PROJECT_NAMESPACE,
                message=err_msg,
                reason="Invalid",
            )

    @patch(
        "toddlers.plugins.unretire_packages.UnretirePackages."
        "_verify_package_not_retired_for_reason"
    )
    def test_is_package_ready_for_unretirement_date_issue(self, verify_reason_mock):
        """
        Assert that package won't be verified if retire date is not correct.
        """
        issue_id = 55
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "242142144"
        err_msg = "Couldn't get a date of the retire commit."

        with patch(
            "toddlers.plugins.unretire_packages.UnretirePackages"
            "._verify_bugzilla_ticket",
            side_effect=ValidationError(err_msg),
        ):
            assert (
                self.toddler._is_package_ready_for_unretirement(
                    issue_id, branches, review_bugzilla
                )
                is False
            )
            self.toddler.pagure_io.close_issue.assert_called_with(
                issue_id=issue_id,
                namespace=unretire_packages.PROJECT_NAMESPACE,
                message=err_msg,
                reason="Invalid",
            )


class TestVerifyPackageNotRetiredForReason:
    """
    Test class for `toddlers.plugin.unretire_packages.UnreturePackages
    ._verify_package_not_retired_for_reason` method.
    """

    def setup_method(self):
        """
        Initialize toddler.
        """
        self.toddler = unretire_packages.UnretirePackages()
        self.toddler.git_repo = Mock()

    def test_verify_package_not_retired_for_reason_success(self):
        """
        Assert that method process correctly if commit message doesn't contain forbidden keyword.
        """
        branches = ["rawhide", "some_branch"]
        self.toddler.git_repo.get_last_commit_message.return_value = (
            "everything ok issue"
        )

        self.toddler._verify_package_not_retired_for_reason(branches)

    def test_verify_package_not_retired_for_reason_reason_exist(self):
        """
        Assert that method raise Exception if commit message contain forbidden keyword.
        """
        err_msg = "Package was retired for a reason: legal or license issue."
        branches = ["rawhide", "some_branch"]

        self.toddler.git_repo.get_last_commit_message.return_value = "legal issue"

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_package_not_retired_for_reason(branches)


class TestVerifyBugzillaTicket:
    """
    Test class for `toddlers.plugin.unretire_packages.UnreturePackages
    ._verify_bugzilla_ticket` method.
    """

    def setup_method(self):
        """
        Initialize toddler.
        """
        self.toddler = unretire_packages.UnretirePackages()
        self.toddler.git_repo = Mock()

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_date_is_none(self, get_bug_mock):
        """
        Assert that method will raise Exception if git won't be able to find a date.
        """
        err_msg = "Couldn't get a date of the retire commit."
        branches = ["rawhide", "some_branch"]
        bugzilla_review = "321321312"

        self.toddler.git_repo.get_last_commit_date.return_value = None

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(
                review_bugzilla=bugzilla_review, branches=branches
            )

        get_bug_mock.assert_not_called()

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_no_need_to_test_bz(self, get_bug_mock):
        review_bugzilla = ""
        branches = ["rawhide", "some_branch"]
        commit_date = arrow.utcnow().shift(days=-3)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)
        get_bug_mock.assert_not_called()

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_bugzilla_id_is_missing(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        review_bugzilla = ""
        branches = ["rawhide", "some_branch"]
        err_msg = "Bugzilla url is missing, please add it and recreate the ticket."
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)

    @patch(
        "toddlers.plugins.unretire_packages.bugzilla_system.get_bug",
        side_effect=Exception("bug error"),
    )
    def test_verify_bugzilla_ticket_bug_getting_error(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "lalalala"
        err_msg = (
            "The Bugzilla bug could not be verified. The following error was encountered: "
            "bug error"
        )
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)

    @patch(
        "toddlers.plugins.unretire_packages.bugzilla_system.get_bug",
        return_value=None,
    )
    def test_verify_bugzilla_ticket_bug_is_none(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "lalalala"
        err_msg = "Bugzilla can't get the bug by bug id, fix bugzilla url."
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_bug_product_is_not_fedora(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "321321321321"
        bug = MagicMock()
        bug.product = "Some product"
        err_msg = (
            "The bugzilla bug is for '{0}', but request should be for 'Fedora'.".format(
                bug.product
            )
        )
        get_bug_mock.return_value = bug
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_bug_does_not_have_required_flag(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "32131321"
        err_msg = (
            "Tag fedora-review is missing on bugzilla, get it and recreate the ticket."
        )

        bug = MagicMock()
        bug.product = "Fedora"
        bug.get_flags.side_effect = TypeError
        get_bug_mock.return_value = bug
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_wrong_flag_status(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "3213213213"
        err_msg = "Flag fedora-review has wrong status, need to be +"
        bug = MagicMock()
        bug.product = "Fedora"
        bug.get_flags.return_value = [
            {
                "status": "-",
            },
        ]
        get_bug_mock.return_value = bug
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)

    @patch("toddlers.plugins.unretire_packages.bugzilla_system.get_bug")
    def test_verify_bugzilla_ticket_success(self, get_bug_mock):
        """
        Assert that method will raise Exception if url is missing in issue body.
        """
        branches = ["rawhide", "some_branch"]
        review_bugzilla = "3213213213"
        bug = MagicMock()
        bug.product = "Fedora"
        bug.get_flags.return_value = [
            {
                "status": "+",
            },
        ]
        get_bug_mock.return_value = bug
        commit_date = arrow.utcnow().shift(days=-100)
        arrow.get = MagicMock(return_value=commit_date)
        self.toddler.git_repo.get_last_commit_date.return_value = commit_date

        self.toddler._verify_bugzilla_ticket(review_bugzilla, branches)


class TestIsNeedToUnblockTagsOnKoji:
    """
    Test class for `toddlers.plugin.unretire_packages.UnreturePackages
    ._is_need_to_unblock_tags_on_koji` method.
    """

    def setup_method(self):
        """
        Initialize toddler.
        """
        self.toddler = unretire_packages.UnretirePackages()
        self.toddler.koji_session = Mock()

    def test_is_need_to_unblock_tags_on_koji_package_does_not_have_tags(self):
        """
        Assert error during processing if method cant find tags of the package.
        """
        tags_to_unblock = ["f37"]
        repo = "some name"
        err_msg = "Package doesn't have tags on koji."

        self.toddler.koji_session.listTags.return_value = None

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._check_tags_to_unblock(tags_to_unblock, repo)

    def test_is_need_to_unblock_tags_on_koji_cant_find_the_package(self):
        """
        Assert error during processing if koji not able to find the package.
        """
        tags_to_unblock = ["f37"]
        repo = "some name"
        err_msg = "Package doesn't exist on koji."

        self.toddler.koji_session.listTags.side_effect = koji.GenericError

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._check_tags_to_unblock(tags_to_unblock, repo)

    def test_is_need_to_unblock_tags_on_koji_request_tags_that_are_not_exist_on_koji(
        self,
    ):
        """
        Assert error during processing, tags are requested to unblock are not exist on koji.
        """
        tags_to_unblock = ["f37"]
        repo = "some name"
        err_msg = "Request to unblock tags that don't exist on koji."

        self.toddler.koji_session.listTags.return_value = [
            {
                "id": 1,
                "locked": True,
                "name": "dist-f33",
            },
            {
                "id": 2,
                "locked": False,
                "name": "dist-f34",
            },
        ]

        with pytest.raises(ValidationError, match=err_msg):
            self.toddler._check_tags_to_unblock(tags_to_unblock, repo)

    def test_is_need_to_unblock_tags_on_koji_true(self):
        """
        Assert that is need to unblock method process correctly and will return right value.
        """
        tags_to_unblock = ["f37"]
        repo = "some name"

        self.toddler.koji_session.listTags.return_value = [
            {
                "id": 1,
                "locked": True,
                "name": "dist-f37",
            },
            {
                "id": 2,
                "locked": False,
                "name": "dist-f34",
            },
        ]

        assert self.toddler._check_tags_to_unblock(tags_to_unblock, repo) is True

    def test_is_need_to_unblock_tags_on_koji_false(self):
        """
        Assert that is need to unblock method process correctly and will return right value.
        """
        tags_to_unblock = ["f37"]
        repo = "some name"

        self.toddler.koji_session.listTags.return_value = [
            {
                "id": 1,
                "locked": False,
                "name": "dist-f37",
            },
            {
                "id": 2,
                "locked": False,
                "name": "dist-f34",
            },
        ]

        assert self.toddler._check_tags_to_unblock(tags_to_unblock, repo) is False


class TestNsConvertor:
    """
    Test class for `toddlers.plugins.unretire_packages.UnretirePackages.
    _ns_convertor` method.`
    """

    def setup_method(self):
        self.toddler = unretire_packages.UnretirePackages()

    def test_ns_convertor(self):
        assert self.toddler._ns_convertor(None) is None
        assert self.toddler._ns_convertor("") == ""
        assert self.toddler._ns_convertor("rpm") == "rpms"
        assert self.toddler._ns_convertor("test") == "tests"
        assert self.toddler._ns_convertor("flatpak") == "flatpaks"
        assert self.toddler._ns_convertor("module") == "modules"
        assert self.toddler._ns_convertor("rpms") == "rpms"


class TestIsUrlExist:
    """
    Test class for `toddlers.plugins.unretire_packages.UnretirePackages.
    _is_url_exist` method.`
    """

    def setup_method(self):
        """
        Initialize toddler.
        """
        self.toddler = unretire_packages.UnretirePackages()
        self.toddler.requests_session = MagicMock()

    def test_is_url_exist_connection_error(self):
        """
        Assert that when Connection error appear function will return False
        """
        url = "some url"
        self.toddler.requests_session.get = MagicMock(side_effect=ConnectionError)
        assert self.toddler._does_url_exist(url) is False

    def test_is_url_exist_response_code_okay(self):
        """
        Assert that when function run correctly, True is returned.
        """
        url = "some url"
        self.toddler.requests_session.get = MagicMock()
        self.toddler.requests_session.get.return_value.status_code = 200
        assert self.toddler._does_url_exist(url) is True

    def test_is_url_exist_response_code_not_okay(self):
        """
        Assert that when status of request diff from 200, False is returned.
        """
        url = "some url"
        self.toddler.requests_session.get = MagicMock()
        self.toddler.requests_session.get.return_value.status_code = 404
        assert self.toddler._does_url_exist(url) is False


class TestMain:
    """
    Test class for `toddlers.plugins.unretire_packages.main` function.
    """

    def test_main_no_args(self, capsys):
        """Assert that help is printed if no arg is provided."""
        with pytest.raises(SystemExit):
            unretire_packages.main([])

        out, err = capsys.readouterr()
        assert out == ""
        # Expecting something along these lines, but don't make the test too tight:
        #
        # usage: pytest [-h] [--dry-run] [-q | --debug] conf [username]
        # pytest: error: the following arguments are required: conf
        assert err.startswith("usage:")
        assert "error: the following arguments are required:" in err

    @patch("tomllib.load")
    @patch("toddlers.plugins.unretire_packages.UnretirePackages.process")
    @patch("toddlers.plugins.unretire_packages.pagure")
    def test_main(self, mock_pagure, mock_process, mock_toml):
        """Assert that main is initializing config and message correctly."""
        # Preparation
        mock_toml.return_value = {}
        mock_pagure_io = Mock()
        mock_issue = {"id": 100, "user": {"name": "amedvede"}}
        mock_pagure_io.get_issue.return_value = mock_issue

        mock_pagure.set_pagure.return_value = mock_pagure_io

        # Function to test
        unretire_packages.main(["--config", "test.cfg", "100"])

        # Assertions
        mock_toml.assert_called_with("test.cfg")
        mock_pagure.set_pagure.assert_called_with({})
        mock_pagure_io.get_issue.assert_called_with(
            100, unretire_packages.PROJECT_NAMESPACE
        )
        body = {
            "project": {"fullname": unretire_packages.PROJECT_NAMESPACE},
            "issue": mock_issue,
        }
        message = IssueNewV1(body=body)
        mock_process.assert_called_with(config={}, message=message)
