import logging
from unittest.mock import Mock, patch

import fedora_messaging.api
import fedora_messaging.exceptions
import pytest

import toddlers.runner
from toddlers.utils.misc import merge_dicts


def patch_messaging_config(*add_dicts):
    mock_config = {
        "consumer_config": {"blocked_toddlers": ["debug"], "toddlers": {}},
    }
    for add_dict in add_dicts:
        mock_config = merge_dicts(mock_config, add_dict)

    return patch.dict("toddlers.runner.fedora_messaging.config.conf", mock_config)


class TestRunningToddler:
    @patch_messaging_config()
    def test___init__(self):
        runner = toddlers.runner.RunningToddler()
        # Check that at least these toddlers exist (but don't fail
        # automatically just because new ones are added).
        assert {
            "flag_ci_pr",
            "flag_commit_build",
            "packager_bugzilla_sync",
        } <= {t.name for t in runner.toddlers}

    @patch_messaging_config()
    @patch("toddlers.runner.ToddlerBase")
    def test___init__no_toddlers(self, mock_base, caplog):
        mock_base.__subclasses__ = Mock(return_value=[])
        caplog.set_level(logging.INFO)
        with pytest.raises(
            fedora_messaging.exceptions.ConfigurationException,
            match=r".* No toddlers found to run .*",
        ):
            toddlers.runner.RunningToddler()
        assert caplog.records[-1].message == "Loaded: []"

    @patch("toddlers.runner.ToddlerBase")
    def test___init__two_toddlers(self, base_tod, caplog):
        caplog.set_level(logging.INFO)
        base_tod.__subclasses__ = Mock()
        toddler_list = []
        for name in ["flag_ci_pr", "debug", "flag_ci_pr"]:
            tod = Mock()
            tod.name = name
            tod.return_value = tod
            toddler_list.append(tod)
        base_tod.__subclasses__.return_value = toddler_list
        with pytest.raises(
            ValueError,
            match=r"Toddler named flag_ci_pr already exists.",
        ):
            toddlers.runner.RunningToddler()

    @patch_messaging_config(
        {"consumer_config": {"blocked_toddlers": ["debug", "flag_ci_pr"]}}
    )
    def test___init__blockedlist(self, caplog):
        blocked_toddlers = {"debug", "flag_ci_pr"}
        expected_toddlers = {
            "flag_commit_build",
            "packager_bugzilla_sync",
        }

        runner = toddlers.runner.RunningToddler()
        running_toddlers = {t.name for t in runner.toddlers}

        assert not running_toddlers & blocked_toddlers
        assert running_toddlers & expected_toddlers == expected_toddlers

    @patch_messaging_config({"consumer_config": {"blocked_toddlers": []}})
    def test___call__debug(self, caplog):
        caplog.set_level(logging.INFO)
        msg = fedora_messaging.api.Message()
        msg.id = 123
        msg.topic = "foo.bar"
        msg.body = {
            "owner": "username",
            "instance": "secondary",
        }
        runner = toddlers.runner.RunningToddler()
        runner.__call__(msg)
        assert any(m.startswith("Topics of interest:") for m in caplog.messages)
        assert "Toddlers processing: 123 -- foo.bar" in caplog.messages
        assert (
            caplog.records[-1].message
            == "Toddler 'debug' accepted to process message id: 123"
        )

    @patch_messaging_config({"consumer_config": {"blocked_toddlers": []}})
    @patch(
        "toddlers.plugins.debug.DebugToddler.process",
        new=Mock(side_effect=Exception("haha")),
    )
    def test___call__exception(self, caplog):
        caplog.set_level(logging.INFO)
        msg = fedora_messaging.api.Message()
        msg.id = 123
        msg.topic = "foo.bar"
        msg.body = {
            "owner": "username",
            "instance": "secondary",
        }
        runner = toddlers.runner.RunningToddler()
        with pytest.raises(
            fedora_messaging.exceptions.Nack, match=r"A toddler tripped"
        ):
            runner.__call__(msg)
        assert any(m.startswith("Topics of interest:") for m in caplog.messages)
        assert "Toddlers processing: 123 -- foo.bar" in caplog.messages
        assert "Toddler 'debug' accepted to process message id: 123" in caplog.messages
        assert (
            caplog.records[-1].message
            == "Toddler 'debug' failed to process message id: 123 -- putting it back in the queue"
        )
