"""
Unit tests for `toddlers.utils.anitya`.
"""

from unittest.mock import Mock

import pytest

import toddlers.utils.anitya as anitya


class TestAnityaSetAnitya:
    """
    Test class for `toddlers.anitya.set_anitya` function.
    """

    def test_set_anitya(self):
        """
        Test initialization of anitya module.
        """
        config = {
            "anitya_endpoint": "https://release-monitoring.org",
            "anitya_access_token": "TOKEN",
        }
        anitya_obj = anitya.set_anitya(config)

        assert anitya_obj._anitya_endpoint == config.get("anitya_endpoint")
        assert anitya_obj._anitya_token == config.get("anitya_access_token")
        assert anitya_obj._requests_session

    def test_set_anitya_no_anitya_url(self):
        """
        Test initialization of anitya module without required config value.
        """
        with pytest.raises(
            ValueError, match=r"No anitya endpoint found in config file"
        ):
            anitya.set_anitya({})

    def test_set_anitya_no_anitya_api_key(self):
        """
        Test initialization of anitya module without required config value.
        """
        with pytest.raises(
            ValueError, match=r"No anitya access token found in config file"
        ):
            config = {"anitya_endpoint": "https://anitya.io"}
            anitya.set_anitya(config)


class TestAnityaDoesProjectExistInAnitya:
    """
    Test class for
    `toddlers.anitya.Anitya.does_project_exists_in_anitya` method.
    """

    def setup_method(self):
        """
        Setup method for test class.
        """
        config = {
            "anitya_endpoint": "https://release-monitoring.org",
            "anitya_access_token": "TOKEN",
        }
        self.anitya_obj = anitya.set_anitya(config)
        self.anitya_obj._requests_session = Mock()
        self.anitya_obj.remove_trailing_slashes_from_url = Mock(
            return_value="https://release-monitoring.org/api/v2/projects/"
        )

    def test_does_project_exists_in_anitya(self):
        """
        Assert that method will return correct response about project exists in anitya.
        """
        endpoint = "https://release-monitoring.org/api/v2/projects/"
        project_name = "amedvede_project"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "items": [
                {
                    "id": 123,
                    "name": project_name,
                }
            ],
            "total_items": 1,
        }
        params = {"name": project_name}
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_project_exists_in_anitya(project_name)

        assert result == "https://release-monitoring.org/project/123"
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )

    def test_does_project_exists_in_anitya_project_not_found(self):
        """
        Assert that method will return correct response about project does not exist in anitya.
        """
        endpoint = "https://release-monitoring.org/api/v2/projects/"
        project_name = "amedvede_project"
        mock_response = Mock()
        mock_response.status_code = 404
        params = {"name": project_name}
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_project_exists_in_anitya(project_name)

        assert result is None
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )

    def test_does_project_exists_in_anitya_empty_items(self):
        """Assert that method will return correct response about project not found in anitya."""
        endpoint = "https://release-monitoring.org/api/v2/projects/"
        project_name = "amedvede_project"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"items": [], "total_items": 0}
        params = {"name": project_name}
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_project_exists_in_anitya(project_name)

        assert result is None
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )

    def test_does_project_exists_in_anitya_wrong_structure(self):
        """
        Assert that method will return correct response about project has wrong structure in anitya.
        """
        endpoint = "https://release-monitoring.org/api/v2/projects/"
        project_name = "amedvede_project"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "items": [
                {
                    "wrong": "structure",
                }
            ],
            "total_items": 1,
        }
        params = {"name": project_name}
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_project_exists_in_anitya(project_name)

        assert result is None
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )


class TestAnityaDoesPackageExistInAnitya:
    """
    Test class for `toddlers.anitya.Anitya.does_package_exists_in_anitya` method.
    """

    def setup_method(self):
        """
        Setup method for test class.
        """
        config = {
            "anitya_endpoint": "https://release-monitoring.org",
            "anitya_access_token": "TOKEN",
        }
        self.anitya_obj = anitya.set_anitya(config)
        self.anitya_obj._requests_session = Mock()
        self.anitya_obj.remove_trailing_slashes_from_url = Mock(
            return_value="https://release-monitoring.org/api/v2/packages/"
        )

    @pytest.mark.parametrize(
        "project_name, expected_project_name, expected_result",
        [
            ("nice_project", "nice_project", True),
            ("nice_project", "bad_project", False),
        ],
    )
    def test_does_package_exists_in_anitya(
        self, project_name, expected_project_name, expected_result
    ):
        """
        Assert that method will return correct response about package exists in anitya
        and his project name is same with expected.
        """
        endpoint = "https://release-monitoring.org/api/v2/packages/"
        package_name = "amedvede_package"
        distribution = "Fedora"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "items": [
                {
                    "name": package_name,
                    "project": project_name,
                }
            ],
            "total_items": 1,
        }
        params = {
            "name": package_name,
            "distribution": distribution,
        }
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_package_exists_in_anitya(
            package_name, distribution, expected_project_name
        )

        assert result is expected_result  # package and project name the same
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )

    def test_does_package_exists_in_anitya_not_found(self):
        """
        Assert that method will return correct response when package does not exist in anitya.
        """
        endpoint = "https://release-monitoring.org/api/v2/packages/"
        package_name = "amedvede_package"
        project_name = "different name"
        distribution = "Fedora"
        mock_response = Mock()
        mock_response.status_code = 202
        params = {
            "name": package_name,
            "distribution": distribution,
        }
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_package_exists_in_anitya(
            package_name, distribution, project_name
        )

        assert result is False  # package and project name are different
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )

    def test_does_package_exists_in_anitya_found_zero_items(self):
        """
        Assert that method will return correct response when response code is correct,
        but response does not contain items.
        """
        endpoint = "https://release-monitoring.org/api/v2/packages/"
        package_name = "amedvede_package"
        project_name = "different name"
        distribution = "Fedora"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"items": [], "total_items": 0}
        params = {
            "name": package_name,
            "distribution": distribution,
        }
        self.anitya_obj._requests_session.get.return_value = mock_response

        result = self.anitya_obj.does_package_exists_in_anitya(
            package_name, distribution, project_name
        )

        assert result is False  # package and project name are different
        self.anitya_obj._requests_session.get.assert_called_once_with(
            endpoint, params=params
        )


class TestAnityaCreateProjectInAnitya:
    """
    Test class for `toddlers.anitya.Anitya.create_project_in_anitya` method.
    """

    def setup_method(self):
        """
        Setup method for test class.
        """
        config = {
            "anitya_endpoint": "https://release-monitoring.org",
            "anitya_access_token": "TOKEN",
        }
        self.anitya_obj = anitya.set_anitya(config)
        self.anitya_obj._requests_session = Mock()
        self.anitya_obj.remove_trailing_slashes_from_url = Mock(
            return_value="https://release-monitoring.org/api/v2/projects/"
        )

    def test_create_project_in_anitya_successful_creation(self):
        """
        Assert that method will return correct response when project is created.
        """
        endpoint = "https://release-monitoring.org/api/v2/projects/"
        project_name = "project"
        homepage = "https://project.com"
        backend = "GitHub"
        test_data = {
            "name": project_name,
            "homepage": homepage,
            "backend": backend,
        }
        response_json = {"id": 123}
        mock_response = Mock()
        mock_response.status_code = 201
        mock_response.json.return_value = response_json
        self.anitya_obj._requests_session.post.return_value = mock_response

        result = self.anitya_obj.create_project_in_anitya(
            project_name, homepage, backend
        )

        assert result == "https://release-monitoring.org/project/123"
        self.anitya_obj._requests_session.post.assert_called_once_with(
            url=endpoint,
            data=test_data,
            headers={"Authorization": "token TOKEN"},
        )

    def test_create_project_in_anitya_fail(self):
        """
        Assert that method will return correct response when project is not created.
        """
        endpoint = "https://release-monitoring.org/api/v2/projects/"
        project_name = "project"
        homepage = "https://project.com"
        backend = "GitHub"
        test_data = {
            "name": project_name,
            "homepage": homepage,
            "backend": backend,
        }
        mock_response = Mock()
        mock_response.status_code = 400
        self.anitya_obj._requests_session.post.return_value = mock_response

        result = self.anitya_obj.create_project_in_anitya(
            project_name, homepage, backend
        )

        assert result is None
        self.anitya_obj._requests_session.post.assert_called_once_with(
            url=endpoint,
            data=test_data,
            headers={"Authorization": "token TOKEN"},
        )


class TestAnityaCreatePackageInAnitya:
    """
    Test class for `toddlers.anitya.Anitya.create_package_in_anitya` method.
    """

    def setup_method(self):
        """
        Setup method for test class.
        """
        config = {
            "anitya_endpoint": "https://release-monitoring.org",
            "anitya_access_token": "TOKEN",
        }
        self.anitya_obj = anitya.set_anitya(config)
        self.anitya_obj._requests_session = Mock()
        self.anitya_obj.remove_trailing_slashes_from_url = Mock(
            return_value="https://release-monitoring.org/api/v2/packages/"
        )

    @pytest.mark.parametrize(
        "response_code, expected_result",
        [
            (201, "Success"),
            (400, "Bad Request, some necessary arguments were not provided."),
            (401, "Unauthorized, access token is incorrect."),
            (409, "Conflict, package already exists."),
            (404, None),
        ],
    )
    def test_create_package_in_anitya(self, response_code, expected_result):
        """
        Assert that method will return correct response when package is created.
        """
        endpoint = "https://release-monitoring.org/api/v2/packages/"
        package_name = "test_package"
        project_name = "test_project"
        distribution = "Fedora"
        project_ecosystem = "https://project.com"
        test_data = {
            "package_name": package_name,
            "project_name": project_name,
            "distribution": distribution,
            "project_ecosystem": project_ecosystem,
        }
        mock_response = Mock()
        mock_response.status_code = response_code
        self.anitya_obj._requests_session.post.return_value = mock_response

        result = self.anitya_obj.create_package_in_anitya(
            package_name, project_name, distribution, project_ecosystem
        )

        assert result == expected_result
        self.anitya_obj._requests_session.post.assert_called_once_with(
            url=endpoint,
            data=test_data,
            headers={"Authorization": "token TOKEN"},
        )
