import logging
from unittest.mock import call, MagicMock, Mock, patch
import xmlrpc.client

import pytest

import toddlers.utils.bugzilla_system


class TestBugzillaSystem:
    def test_set_bz_no_bugzilla_url(self):
        with pytest.raises(
            ValueError, match=r"No bugzilla_url found in the configuration file"
        ):
            toddlers.utils.bugzilla_system.set_bz({})

    def test_set_bz_no_bugzilla_api_key(self):
        with pytest.raises(
            ValueError, match=r"No bugzilla_api_key found in the configuration file"
        ):
            config = {
                "bugzilla_url": "https:bz.example.com",
            }
            toddlers.utils.bugzilla_system.set_bz(config)

    @patch("toddlers.utils.bugzilla_system.Bugzilla")
    def test_set_bz(self, mock_bz):
        mock_bz.return_value = "bugzilla_object"
        config = {
            "bugzilla_url": "https:bz.example.com",
            "bugzilla_api_key": "verysecretapikey",
        }
        output = toddlers.utils.bugzilla_system.set_bz(config)
        mock_bz.assert_called_with(
            url="https:bz.example.com/xmlrpc.cgi",
            api_key="verysecretapikey",
            cookiefile=None,
            tokenfile=None,
        )
        assert output == "bugzilla_object"

    @patch("toddlers.utils.bugzilla_system._BUGZILLA", new=None)
    def test_get_bz_not_set(self):
        with pytest.raises(
            ValueError, match=r"No bugzilla connection set, call set_bz first"
        ):
            toddlers.utils.bugzilla_system.get_bz()

    def test_get_bz(self):
        output = toddlers.utils.bugzilla_system.get_bz()
        assert output == "bugzilla_object"

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_group_member(self, mock_bz):
        groups = Mock()
        groups.member_emails = ["foo@bar.com", "foo@baz.com"]
        server = Mock()
        server.getgroup.return_value = groups
        mock_bz.return_value = server

        output = toddlers.utils.bugzilla_system.get_group_member("test")
        mock_bz.assert_called_with()
        server.getgroup.assert_called_with("test", membership=True)
        assert output == ["foo@bar.com", "foo@baz.com"]

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_remove_user_from_group(self, mock_bz):
        server = Mock()
        server.updateperms.return_value = True
        mock_bz.return_value = server

        toddlers.utils.bugzilla_system.remove_user_from_group(
            "test@foo.com", "groupname"
        )
        mock_bz.assert_called_with()
        server.updateperms.assert_called_with("test@foo.com", "rem", "groupname")

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_remove_user_from_group_failed(self, mock_bz):
        server = Mock()
        server.updateperms.side_effect = xmlrpc.client.Fault(55, "error string")
        mock_bz.return_value = server

        with pytest.raises(xmlrpc.client.Fault, match=r"<Fault 55: 'error string'>"):
            toddlers.utils.bugzilla_system.remove_user_from_group(
                "test@foo.com", "groupname"
            )
        mock_bz.assert_called_with()
        server.updateperms.assert_called_with("test@foo.com", "rem", "groupname")

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_remove_user_from_group_no_user(self, mock_bz):
        server = Mock()
        server.updateperms.side_effect = xmlrpc.client.Fault(51, "No such user")
        mock_bz.return_value = server

        toddlers.utils.bugzilla_system.remove_user_from_group(
            "test@foo.com", "groupname"
        )
        mock_bz.assert_called_with()
        server.updateperms.assert_called_with("test@foo.com", "rem", "groupname")

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_user_exists(self, mock_bz):
        server = Mock()
        server.getuser.return_value = True
        mock_bz.return_value = server

        output = toddlers.utils.bugzilla_system.get_user("test@foo.com")
        mock_bz.assert_called_with()
        server.getuser.assert_called_with("test@foo.com")
        assert output

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_user_exists_no_user(self, mock_bz):
        server = Mock()
        server.getuser.side_effect = xmlrpc.client.Fault(51, "No such user")
        mock_bz.return_value = server

        output = toddlers.utils.bugzilla_system.get_user("test@foo.com")
        mock_bz.assert_called_with()
        server.getuser.assert_called_with("test@foo.com")
        assert output is None

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_user_exists_failed(self, mock_bz):
        server = Mock()
        server.getuser.side_effect = xmlrpc.client.Fault(55, "error string")
        mock_bz.return_value = server

        with pytest.raises(xmlrpc.client.Fault, match=r"<Fault 55: 'error string'>"):
            toddlers.utils.bugzilla_system.get_user("test@foo.com")
        mock_bz.assert_called_with()
        server.getuser.assert_called_with("test@foo.com")

    @patch("toddlers.utils.bugzilla_system.get_user", new=Mock(return_value=False))
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_user_to_group_no_user(self, mock_bz):
        server = Mock()
        server.updateperms.return_value = True
        mock_bz.return_value = server

        output = toddlers.utils.bugzilla_system.add_user_to_group(
            user_email="test@foo.com",
            bz_group="groupname",
            no_bz_account=[],
            dry_run=False,
        )
        mock_bz.assert_called_with()
        server.updateperms.assert_not_called()
        assert output == ["test@foo.com"]

    @patch("toddlers.utils.bugzilla_system.get_user")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_user_to_group_dry_run(self, mock_bz, get_user, caplog):
        caplog.set_level(logging.INFO)
        server = Mock()
        server.updateperms.return_value = True
        mock_bz.return_value = server
        user = Mock()
        user.can_login = True
        get_user.return_value = user

        output = toddlers.utils.bugzilla_system.add_user_to_group(
            user_email="test@foo.com",
            bz_group="groupname",
            no_bz_account=[],
            dry_run=True,
        )
        mock_bz.assert_called_with()
        server.updateperms.assert_not_called()
        assert output == []
        assert (
            caplog.records[-1].message
            == "   Would add test@foo.com to the group groupname"
        )

    @patch("toddlers.utils.bugzilla_system.get_user")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_user_to_group(self, mock_bz, get_user, caplog):
        caplog.set_level(logging.INFO)
        server = Mock()
        server.updateperms.return_value = True
        mock_bz.return_value = server
        user = Mock()
        user.can_login = True
        get_user.return_value = user

        output = toddlers.utils.bugzilla_system.add_user_to_group(
            user_email="test@foo.com",
            bz_group="groupname",
            no_bz_account=[],
            dry_run=False,
        )
        mock_bz.assert_called_with()
        server.updateperms.assert_called_with("test@foo.com", "add", "groupname")
        assert output == []

    @patch("toddlers.utils.bugzilla_system.get_user")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_user_to_group_cant_login(self, mock_bz, get_user, caplog):
        caplog.set_level(logging.INFO)
        server = Mock()
        server.updateperms.return_value = True
        mock_bz.return_value = server
        user = Mock()
        user.can_login = False
        get_user.return_value = user

        output = toddlers.utils.bugzilla_system.add_user_to_group(
            user_email="test@foo.com",
            bz_group="groupname",
            no_bz_account=[],
            dry_run=False,
        )
        mock_bz.assert_called_with()
        server.updateperms.assert_not_called()
        assert output == []

    @patch("toddlers.utils.bugzilla_system.get_user")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_user_to_group_user_disabled(self, mock_bz, get_user, caplog):
        caplog.set_level(logging.INFO)
        server = Mock()
        server.updateperms.side_effect = xmlrpc.client.Fault(32000, "Fault")
        mock_bz.return_value = server
        user = Mock()
        user.can_login = True
        get_user.return_value = user

        output = toddlers.utils.bugzilla_system.add_user_to_group(
            user_email="test@foo.com",
            bz_group="groupname",
            no_bz_account=[],
            dry_run=False,
        )

        mock_bz.assert_called_with()
        server.updateperms.assert_called_with("test@foo.com", "add", "groupname")
        assert output == []
        assert (
            caplog.records[-1].message
            == "   User test@foo.com is missing in LDAP. Skipping."
        )


class TestGetProductInfoPackager:
    """Test class for `toddlers.utils.bugzilla_system.get_product_info_packages` function."""

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_product_info_packages(self, mock_bz):
        """Assert that product info is received."""
        server = MagicMock()
        server.product_get.return_value = [
            {
                "components": [
                    {
                        "name": "foo",
                        "default_assigned_to": "default_assignee",
                        "description": "description",
                        "default_qa_contact": "default_qa_contact",
                        "default_cc": "default_cc",
                        "is_active": True,
                    }
                ]
            }
        ]
        mock_bz.return_value = server

        output = toddlers.utils.bugzilla_system.get_product_info_packages(
            collection="Fedora",
        )

        mock_bz.assert_called_with()
        server.product_get.assert_called_with(
            names=["Fedora"],
            include_fields=[
                "components.name",
                "components.default_assigned_to",
                "components.description",
                "components.default_qa_contact",
                "components.default_cc",
                "components.is_active",
            ],
        )

        assert output == {
            "foo": {
                "initialowner": "default_assignee",
                "description": "description",
                "initialqacontact": "default_qa_contact",
                "initialcclist": "default_cc",
                "is_active": True,
            }
        }

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_product_info_packages_exception(self, mock_bz):
        """Assert that exception is raised when data are not in correct format."""
        server = MagicMock()
        server.product_get.return_value = []
        mock_bz.return_value = server

        with pytest.raises(ValueError) as exc:
            toddlers.utils.bugzilla_system.get_product_info_packages(
                collection="Fedora",
            )

        mock_bz.assert_called_with()
        server.product_get.assert_called_with(
            names=["Fedora"],
            include_fields=[
                "components.name",
                "components.default_assigned_to",
                "components.description",
                "components.default_qa_contact",
                "components.default_cc",
                "components.is_active",
            ],
        )
        assert str(exc.value) == "Received data doesn't have the expected format"


class TestReassignTicketsToAssignee:
    """Test class for `toddlers.utils.bugzilla_system.reassign_tickets_to_assignee` function."""

    # Query used in multiple tests
    query = {
        "product": "Fedora",
        "component": "foo",
        "bug_status": [
            "NEW",
            "ASSIGNED",
            "ON_DEV",
            "ON_QA",
            "MODIFIED",
            "POST",
            "FAILS_QA",
            "PASSES_QA",
            "RELEASE_PENDING",
        ],
        "version": ["rpms"],
    }

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_reassign_tickets_to_assignee_dry_run(self, mock_bz):
        """Assert that dry_run doesn't do any change."""
        old_poc = "khorne@fedoraproject.org"
        new_poc = "tzeentch@fedoraproject.org"
        bug = Mock()
        bug.assigned_to = old_poc

        server = Mock()
        server.query.return_value = [bug]

        mock_bz.return_value = server

        toddlers.utils.bugzilla_system.reassign_tickets_to_assignee(
            new_poc=new_poc,
            old_poc=old_poc,
            product="Fedora",
            package="foo",
            versions=["rpms"],
            dry_run=True,
        )

        mock_bz.assert_called_with()
        server.query.assert_called_with(query=self.query)
        bug.setassignee.asssert_not_called()

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_reassign_tickets_to_assignee_print_fas_names_no_fas_info(
        self, mock_bz, caplog
    ):
        """Assert that `print_fas_names` parameter is handled correctly
        when the FAS user info is missing.
        """
        old_poc = "khorne@fedoraproject.org"
        new_poc = "tzeentch@fedoraproject.org"
        bug = Mock()
        bug.bug_id = 1
        bug.assigned_to = old_poc

        server = Mock()
        server.query.return_value = [bug]

        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.reassign_tickets_to_assignee(
                new_poc=new_poc,
                old_poc=old_poc,
                product="Fedora",
                package="foo",
                versions=["rpms"],
                dry_run=True,
                print_fas_names=True,
            )

        assert (
            "Fedora/foo reassigning bug #1 from khorne@... to tzeentch@..."
            in caplog.text
        )

        mock_bz.assert_called_with()
        server.query.assert_called_with(query=self.query)
        bug.setassignee.asssert_not_called()

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_reassign_tickets_to_assignee_print_fas_names_fas_info(
        self, mock_bz, caplog
    ):
        """Assert that `print_fas_names` parameter is handled correctly
        when the FAS user info is available.
        """
        old_poc = "khorne@fedoraproject.org"
        new_poc = "tzeentch@fedoraproject.org"
        bug = Mock()
        bug.bug_id = 1
        bug.assigned_to = old_poc

        server = Mock()
        server.query.return_value = [bug]

        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.reassign_tickets_to_assignee(
                new_poc=new_poc,
                old_poc=old_poc,
                product="Fedora",
                package="foo",
                versions=["rpms"],
                fas_users_info={
                    "khorne@fedoraproject.org": "Blood God",
                    "tzeentch@fedoraproject.org": "Master of Fate",
                },
                dry_run=True,
                print_fas_names=True,
            )

        assert (
            "Fedora/foo reassigning bug #1 from Blood God to Master of Fate"
            in caplog.text
        )

        mock_bz.assert_called_with()
        server.query.assert_called_with(query=self.query)
        bug.setassignee.asssert_not_called()

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_reassign_tickets_to_assignee_xmlrpc_Fault(self, mock_bz, mock_bz_call):
        """Assert that `xmlrpc.client.Fault` is raised correctly."""
        old_poc = "khorne@fedoraproject.org"
        new_poc = "tzeentch@fedoraproject.org"
        bug = Mock()
        bug.bug_id = 1
        bug.assigned_to = old_poc

        server = Mock()

        mock_bz.return_value = server
        mock_bz_call.side_effect = ([bug], xmlrpc.client.Fault(50, "Fault"))

        with pytest.raises(xmlrpc.client.Fault) as exc_info:
            toddlers.utils.bugzilla_system.reassign_tickets_to_assignee(
                new_poc=new_poc,
                old_poc=old_poc,
                product="Fedora",
                package="foo",
                versions=["rpms"],
            )

        assert exc_info.value.args == ("tzeentch@fedoraproject.org", 50, "Fault")

        mock_bz.assert_called_with()
        mock_bz_call.assert_has_calls(
            [
                call(server.query, kwargs={"query": self.query}),
                call(
                    bug.setassignee,
                    kwargs={
                        "assigned_to": new_poc,
                        "comment": "This package has changed maintainer in Fedora. "
                        + "Reassigning to the new maintainer of this component.",
                    },
                ),
            ]
        )

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_reassign_tickets_to_assignee_xmlrpc_ProtocolError(
        self, mock_bz, mock_bz_call
    ):
        """Assert that `xmlrpc.client.ProtocolError` is raised correctly."""
        old_poc = "khorne@fedoraproject.org"
        new_poc = "tzeentch@fedoraproject.org"
        bug = Mock()
        bug.bug_id = 1
        bug.assigned_to = old_poc

        server = Mock()

        mock_bz.return_value = server
        mock_bz_call.side_effect = (
            [bug],
            xmlrpc.client.ProtocolError("Error", 10, "Error message", {}),
        )

        with pytest.raises(xmlrpc.client.ProtocolError) as exc_info:
            toddlers.utils.bugzilla_system.reassign_tickets_to_assignee(
                new_poc=new_poc,
                old_poc=old_poc,
                product="Fedora",
                package="foo",
                versions=["rpms"],
            )

        assert exc_info.value.args == ("ProtocolError", 10, "Error message")

        mock_bz.assert_called_with()
        mock_bz_call.assert_has_calls(
            [
                call(server.query, kwargs={"query": self.query}),
                call(
                    bug.setassignee,
                    kwargs={
                        "assigned_to": new_poc,
                        "comment": "This package has changed maintainer in Fedora. "
                        + "Reassigning to the new maintainer of this component.",
                    },
                ),
            ]
        )


@pytest.fixture
def owner():
    """Fixture for owner value.
    Used in test classes for both add and edit component.
    """
    return "khorne@fedoraproject.org"


@pytest.fixture
def data(owner):
    """Fixture for data value.
    Used in test classes for both add and edit component.
    """
    return {
        "product": "Fedora",
        "component": "foo",
        "initialowner": owner,
        "description": "description",
        "initialqacontact": "nurgle@fedoraproject.org",
        "initialcclist": [owner],
        "is_active": True,
    }


class TestAddComponent:
    """Test class for `toddlers.utils.bugzilla_system.add_component` function."""

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_component_dry_run(self, mock_bz, caplog, owner):
        """Assert that dry_run doesn't do any change."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.add_component(
                product="Fedora",
                owner=owner,
                package="foo",
                qa_contact="nurgle@fedoraproject.org",
                cc_list=[owner],
                dry_run=True,
            )

        assert (
            caplog.records[0].getMessage()
            == f"[ADDCOMP] Fedora/foo initialowner set to `{owner}`"
        )
        assert (
            caplog.records[1].getMessage()
            == "[ADDCOMP] Fedora/foo description set to `NA`"
        )
        assert (
            caplog.records[2].getMessage()
            == "[ADDCOMP] Fedora/foo initialqacontact set to `nurgle@fedoraproject.org`"
        )
        assert (
            caplog.records[3].getMessage()
            == f"[ADDCOMP] Fedora/foo initialcclist set to `['{owner}']`"
        )
        assert (
            caplog.records[4].getMessage()
            == "[ADDCOMP] Fedora/foo is_active set to `True`"
        )
        server.addcomponent.assert_not_called()

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_component_retired(self, mock_bz, caplog, owner):
        """Assert that nothing is done for retired package."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.add_component(
                product="Fedora",
                owner=owner,
                package="foo",
                qa_contact="nurgle@fedoraproject.org",
                cc_list=[owner],
                retired=True,
                dry_run=True,
            )

        assert caplog.records[0].getMessage() == "[NOADD] Fedora/foo is retired"
        server.addcomponent.assert_not_called()

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_component_print_fas_names(self, mock_bz, caplog, owner, data):
        """Assert that print_fas_names works correctly."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.add_component(
                product="Fedora",
                owner=owner,
                package="foo",
                qa_contact="nurgle@fedoraproject.org",
                cc_list=[owner],
                fas_users_info={
                    "khorne@fedoraproject.org": "Blood God",
                    "nurgle@fedoraproject.org": "Papa Nurgle",
                },
                print_fas_names=True,
            )

        assert (
            caplog.records[0].getMessage()
            == "[ADDCOMP] Fedora/foo initialowner set to FAS name(s) `Blood God`"
        )
        assert (
            caplog.records[1].getMessage()
            == "[ADDCOMP] Fedora/foo description set to `NA`"
        )
        assert (
            caplog.records[2].getMessage()
            == "[ADDCOMP] Fedora/foo initialqacontact set to FAS name(s) `Papa Nurgle`"
        )
        assert (
            caplog.records[3].getMessage()
            == "[ADDCOMP] Fedora/foo initialcclist set to FAS name(s) `['Blood God']`"
        )
        assert (
            caplog.records[4].getMessage()
            == "[ADDCOMP] Fedora/foo is_active set to `True`"
        )

        data["description"] = "NA"

        server.addcomponent.assert_called_with(data=data)

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_add_component_xmlrpc_Fault(self, mock_bz, mock_bz_call, owner, data):
        """Assert that `xmlrpc.client.Fault` is raised correctly."""
        server = Mock()
        mock_bz.return_value = server
        mock_bz_call.side_effect = xmlrpc.client.Fault(50, "Fault")

        with pytest.raises(xmlrpc.client.Fault) as exc_info:
            toddlers.utils.bugzilla_system.add_component(
                product="Fedora",
                owner=owner,
                package="foo",
                qa_contact="nurgle@fedoraproject.org",
                cc_list=[owner],
                fas_users_info={
                    "khorne@fedoraproject.org": "Blood God",
                    "nurgle@fedoraproject.org": "Papa Nurgle",
                },
                print_fas_names=True,
            )

        data["description"] = "NA"

        assert exc_info.value.args == (data, 50, "Fault")

        mock_bz_call.assert_called_with(server.addcomponent, kwargs={"data": data})


class TestEditComponent:
    """Test class for `toddlers.utils.bugzilla_system.edit_component` function."""

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_edit_component_no_change(self, mock_bz, caplog, owner):
        """Assert that nothing is done when no change occurs."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.edit_component(
                owner=owner,
                product="Fedora",
                package="foo",
                component={
                    "initialowner": owner,
                    "description": "description",
                    "initialqacontact": "",
                    "initialcclist": "",
                    "is_active": True,
                },
                cc_list=[],
                versions=[],
                qa_contact="",
                print_no_change=True,
            )

        mock_bz.assert_called_with()
        assert caplog.records[0].getMessage() == "[NOCHANGE] Fedora/foo"

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_edit_component_change_same_owner_dry_run(self, mock_bz, caplog, owner):
        """Assert that component is not updated when change occurs, but dry run is set."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.edit_component(
                owner=owner,
                product="Fedora",
                package="foo",
                component={
                    "initialowner": owner,
                    "description": "",
                    "initialqacontact": "",
                    "initialcclist": [],
                    "is_active": False,
                },
                cc_list=[owner],
                versions=[],
                description="description",
                qa_contact="nurgle@fedoraproject.org",
                dry_run=True,
            )

        mock_bz.assert_called_with()
        server.editcomponent.assert_not_called()
        assert (
            caplog.records[0].getMessage()
            == "[EDITCOMP] Fedora/foo description changed from `` to `description`"
        )
        assert (
            caplog.records[1].getMessage()
            == "[EDITCOMP] Fedora/foo initialqacontact changed from `` "
            "to `nurgle@fedoraproject.org`"
        )
        assert (
            caplog.records[2].getMessage()
            == "[EDITCOMP] Fedora/foo initialcclist changed from `[]` "
            "to `['khorne@fedoraproject.org']`"
        )
        assert (
            caplog.records[3].getMessage()
            == "[EDITCOMP] Fedora/foo is_active changed from `False` to `True`"
        )

    @patch("toddlers.utils.bugzilla_system.reassign_tickets_to_assignee")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_edit_component_change_print_fas_names(
        self, mock_bz, mock_set_pkg_owner, caplog, owner, data
    ):
        """Assert that FAS names are printed when the option is enabled."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.edit_component(
                owner=owner,
                product="Fedora",
                package="foo",
                component={
                    "initialowner": "tzeentch@fedoraproject.org",
                    "description": "",
                    "initialqacontact": "",
                    "initialcclist": ["slaanesh@fedoraproject.org"],
                    "is_active": False,
                },
                cc_list=[owner],
                versions=[],
                description="description",
                qa_contact="nurgle@fedoraproject.org",
                fas_users_info={
                    owner: "Blood God",
                    "nurgle@fedoraproject.org": "Papa Nurgle",
                    "tzeentch@fedoraproject.org": "Master of Fate",
                },
                print_fas_names=True,
            )

        mock_bz.assert_called_with()
        server.editcomponent.assert_called_with(data=data)
        mock_set_pkg_owner.assert_called_with(
            new_poc=owner,
            old_poc="tzeentch@fedoraproject.org",
            product="Fedora",
            package="foo",
            versions=[],
            fas_users_info={
                owner: "Blood God",
                "nurgle@fedoraproject.org": "Papa Nurgle",
                "tzeentch@fedoraproject.org": "Master of Fate",
            },
            dry_run=False,
            print_fas_names=True,
        )
        assert (
            caplog.records[0].getMessage()
            == "[EDITCOMP] Fedora/foo initialowner changed from `Master of Fate` to "
            "FAS name(s) `Blood God`"
        )
        assert (
            caplog.records[1].getMessage()
            == "[EDITCOMP] Fedora/foo description changed from `` to `description`"
        )
        assert (
            caplog.records[2].getMessage()
            == "[EDITCOMP] Fedora/foo initialqacontact changed from `` to FAS name(s) `Papa Nurgle`"
        )
        assert (
            caplog.records[3].getMessage()
            == "[EDITCOMP] Fedora/foo initialcclist changed from "
            "`['slaanesh@fedoraproject.org']` to FAS name(s) `['Blood God']`"
        )
        assert (
            caplog.records[4].getMessage()
            == "[EDITCOMP] Fedora/foo is_active changed from `False` to `True`"
        )

    @patch("toddlers.utils.bugzilla_system.reassign_tickets_to_assignee")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_edit_component_drop_initialqacontact(
        self, mock_bz, mock_set_pkg_owner, caplog, owner, data
    ):
        """Assert that FAS names are printed when the option is enabled."""
        server = Mock()
        mock_bz.return_value = server

        with caplog.at_level(logging.DEBUG):
            toddlers.utils.bugzilla_system.edit_component(
                owner=owner,
                product="Fedora",
                package="foo",
                component={
                    "initialowner": "tzeentch@fedoraproject.org",
                    "description": "",
                    "initialqacontact": "",
                    "initialcclist": ["slaanesh@fedoraproject.org"],
                    "is_active": False,
                },
                cc_list=[owner],
                versions=[],
                description="description",
                qa_contact=None,
                fas_users_info={
                    owner: "Blood God",
                    "nurgle@fedoraproject.org": "Papa Nurgle",
                    "tzeentch@fedoraproject.org": "Master of Fate",
                },
                print_fas_names=True,
            )

        mock_bz.assert_called_with()
        del data["initialqacontact"]
        server.editcomponent.assert_called_with(data=data)
        mock_set_pkg_owner.assert_called_with(
            new_poc=owner,
            old_poc="tzeentch@fedoraproject.org",
            product="Fedora",
            package="foo",
            versions=[],
            fas_users_info={
                owner: "Blood God",
                "nurgle@fedoraproject.org": "Papa Nurgle",
                "tzeentch@fedoraproject.org": "Master of Fate",
            },
            dry_run=False,
            print_fas_names=True,
        )
        assert (
            caplog.records[0].getMessage()
            == "[EDITCOMP] Fedora/foo initialowner changed from `Master of Fate` to "
            "FAS name(s) `Blood God`"
        )
        assert (
            caplog.records[1].getMessage()
            == "[EDITCOMP] Fedora/foo description changed from `` to `description`"
        )
        assert (
            caplog.records[2].getMessage()
            == "[EDITCOMP] Fedora/foo initialqacontact changed from `` to FAS name(s) `None`"
        )
        assert (
            caplog.records[3].getMessage()
            == "[EDITCOMP] Fedora/foo initialcclist changed from "
            "`['slaanesh@fedoraproject.org']` to FAS name(s) `['Blood God']`"
        )
        assert (
            caplog.records[4].getMessage()
            == "[EDITCOMP] Fedora/foo is_active changed from `False` to `True`"
        )

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_edit_component_change_xmlrpc_fault(
        self, mock_bz, mock_bz_call, owner, data
    ):
        """Assert that `xmlrpc.client.Fault` is handled correctly."""
        server = Mock()
        mock_bz.return_value = server
        mock_bz_call.side_effect = xmlrpc.client.Fault(50, "Fault")

        with pytest.raises(xmlrpc.client.Fault) as exc:
            toddlers.utils.bugzilla_system.edit_component(
                owner=owner,
                product="Fedora",
                package="foo",
                component={
                    "initialowner": "tzeentch@fedoraproject.org",
                    "description": "",
                    "initialqacontact": "",
                    "initialcclist": [],
                    "is_active": False,
                },
                cc_list=[owner],
                versions=[],
                description="description",
                qa_contact="nurgle@fedoraproject.org",
                fas_users_info={
                    owner: "Blood God",
                    "nurgle@fedoraproject.org": "Papa Nurgle",
                    "tzeentch@fedoraproject.org": "Master of Fate",
                },
            )

        assert exc.value.args == (data, 50, "Fault")

        mock_bz.assert_called_with()
        mock_bz_call.assert_called_with(server.editcomponent, kwargs={"data": data})

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_edit_component_change_xmlrpc_ProtocolError(
        self, mock_bz, mock_bz_call, owner, data
    ):
        """Assert that `xmlrpc.client.Fault` is handled correctly."""
        server = Mock()
        mock_bz.return_value = server
        mock_bz_call.side_effect = xmlrpc.client.ProtocolError(
            "Error", 10, "Error message", {}
        )

        with pytest.raises(xmlrpc.client.ProtocolError) as exc:
            toddlers.utils.bugzilla_system.edit_component(
                owner=owner,
                product="Fedora",
                package="foo",
                component={
                    "initialowner": "tzeentch@fedoraproject.org",
                    "description": "",
                    "initialqacontact": "",
                    "initialcclist": [],
                    "is_active": False,
                },
                cc_list=[owner],
                versions=[],
                description="description",
                qa_contact="nurgle@fedoraproject.org",
                fas_users_info={
                    owner: "Blood God",
                    "nurgle@fedoraproject.org": "Papa Nurgle",
                    "tzeentch@fedoraproject.org": "Master of Fate",
                },
            )

        assert exc.value.args == ("ProtocolError", 10, "Error message")

        mock_bz.assert_called_with()
        mock_bz_call.assert_called_with(server.editcomponent, kwargs={"data": data})


class TestGetBug:
    """Test class for `toddlers.utils.bugzilla_system.get_bug` function."""

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_bug(self, mock_bz):
        """Assert that bug is retrieved without issue."""
        bug_id = "100"
        server = Mock()
        mock_bug = Mock()
        server.getbug.return_value = mock_bug

        mock_bz.return_value = server

        bug = toddlers.utils.bugzilla_system.get_bug(bug_id)

        mock_bz.assert_called_with()
        server.getbug.assert_called_with(bug_id)
        assert bug == mock_bug

    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_bug_none_found(self, mock_bz):
        """Assert that no bug found returns None."""
        bug_id = "100"
        server = Mock()
        server.getbug.return_value = None

        mock_bz.return_value = server

        bug = toddlers.utils.bugzilla_system.get_bug(bug_id)

        mock_bz.assert_called_with()
        server.getbug.assert_called_with(bug_id)
        assert bug is None

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_bug_xmlrpc_fault(self, mock_bz, mock_bz_call):
        """Assert that XMLRPC fault is handled."""
        bug_id = "100"
        server = Mock()
        mock_bz_call.side_effect = xmlrpc.client.Fault(50, "Fault")

        mock_bz.return_value = server

        with pytest.raises(xmlrpc.client.Fault) as exc:
            toddlers.utils.bugzilla_system.get_bug(bug_id)

        assert exc.value.args == (bug_id, 50, "Fault")

        mock_bz.assert_called_with()
        mock_bz_call.assert_called_with(server.getbug, args=(bug_id,))

    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    @patch("toddlers.utils.bugzilla_system.get_bz")
    def test_get_bug_protocol_error(self, mock_bz, mock_bz_call):
        """Assert that protocol error is handled."""
        bug_id = "100"
        server = Mock()
        mock_bz_call.side_effect = xmlrpc.client.ProtocolError(
            "Error", 10, "Error message", {}
        )

        mock_bz.return_value = server

        with pytest.raises(xmlrpc.client.ProtocolError) as exc:
            toddlers.utils.bugzilla_system.get_bug(bug_id)

        assert exc.value.args == ("ProtocolError", 10, "Error message")

        mock_bz.assert_called_with()
        mock_bz_call.assert_called_with(server.getbug, args=(bug_id,))


class TestCommentOnBug:
    """Test class for `toddlers.utils.bugzilla_system.comment_on_bug` function."""

    @patch("toddlers.utils.bugzilla_system.get_bug")
    def test_comment_on_ticket(self, mock_get_bug):
        """Assert that comment is called without issue."""
        bug_id = "100"
        comment = "comment"

        bug = Mock()

        mock_get_bug.return_value = bug

        toddlers.utils.bugzilla_system.comment_on_bug(bug_id, comment)

        mock_get_bug.assert_called_with(bug_id)
        bug.addcomment.assert_called_with(comment)

    @patch("toddlers.utils.bugzilla_system.get_bug")
    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    def test_comment_on_bug_xmlrpc_fault(self, mock_bz_call, mock_get_bug):
        """Assert that XMLRPC fault is handled."""
        bug_id = "100"
        comment = "comment"

        mock_bz_call.side_effect = xmlrpc.client.Fault(50, "Fault")

        bug = Mock()
        mock_get_bug.return_value = bug

        with pytest.raises(xmlrpc.client.Fault) as exc:
            toddlers.utils.bugzilla_system.comment_on_bug(bug_id, comment)

        assert exc.value.args == (comment, 50, "Fault")

        mock_get_bug.assert_called_with(bug_id)
        mock_bz_call.assert_called_with(bug.addcomment, args=(comment,))

    @patch("toddlers.utils.bugzilla_system.get_bug")
    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    def test_comment_on_bug_protocol_error(self, mock_bz_call, mock_get_bug):
        """Assert that protocol error is handled."""
        bug_id = "100"
        comment = "comment"

        mock_bz_call.side_effect = xmlrpc.client.ProtocolError(
            "Error", 10, "Error message", {}
        )

        bug = Mock()
        mock_get_bug.return_value = bug

        with pytest.raises(xmlrpc.client.ProtocolError) as exc:
            toddlers.utils.bugzilla_system.comment_on_bug(bug_id, comment)

        assert exc.value.args == ("ProtocolError", 10, "Error message")

        mock_get_bug.assert_called_with(bug_id)
        mock_bz_call.assert_called_with(bug.addcomment, args=(comment,))


class TestChangeBugStatus:
    """Test class for `toddlers.utils.bugzilla_system.change_bug_status` function."""

    @patch("toddlers.utils.bugzilla_system.get_bug")
    def test_change_ticket_status(self, mock_get_bug):
        """Assert that change is called without issue."""
        bug_id = "100"
        status = "POST"
        comment = "comment"

        bug = Mock()

        mock_get_bug.return_value = bug

        toddlers.utils.bugzilla_system.change_bug_status(bug_id, status, comment)

        mock_get_bug.assert_called_with(bug_id)
        bug.setstatus.assert_called_with(status, comment)

    @patch("toddlers.utils.bugzilla_system.get_bug")
    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    def test_comment_on_bug_xmlrpc_fault(self, mock_bz_call, mock_get_bug):
        """Assert that XMLRPC fault is handled."""
        bug_id = "100"
        status = "POST"
        comment = "comment"

        mock_bz_call.side_effect = xmlrpc.client.Fault(50, "Fault")

        bug = Mock()
        mock_get_bug.return_value = bug

        with pytest.raises(xmlrpc.client.Fault) as exc:
            toddlers.utils.bugzilla_system.change_bug_status(bug_id, status, comment)

        assert exc.value.args == (comment, 50, "Fault")

        mock_get_bug.assert_called_with(bug_id)
        mock_bz_call.assert_called_with(
            bug.setstatus,
            args=(
                status,
                comment,
            ),
        )

    @patch("toddlers.utils.bugzilla_system.get_bug")
    @patch("toddlers.utils.bugzilla_system.execute_bugzilla_call")
    def test_comment_on_bug_protocol_error(self, mock_bz_call, mock_get_bug):
        """Assert that protocol error is handled."""
        bug_id = "100"
        status = "POST"
        comment = "comment"

        mock_bz_call.side_effect = xmlrpc.client.ProtocolError(
            "Error", 10, "Error message", {}
        )

        bug = Mock()
        mock_get_bug.return_value = bug

        with pytest.raises(xmlrpc.client.ProtocolError) as exc:
            toddlers.utils.bugzilla_system.change_bug_status(bug_id, status, comment)

        assert exc.value.args == ("ProtocolError", 10, "Error message")

        mock_get_bug.assert_called_with(bug_id)
        mock_bz_call.assert_called_with(
            bug.setstatus,
            args=(
                status,
                comment,
            ),
        )


class TestExecuteBugzillaCall:
    """Test class for `toddlers.utils.bugzilla_system.execute_bugzilla_call` function."""

    def test_execute_bugzilla_call(self):
        """Assert that bugzilla call wrapper is working correctly."""
        mock_call = MagicMock()
        mock_call.return_value = {"name": "foo"}
        kwargs = {"product": "Fedora", "data": {"component": "foo"}}

        output = toddlers.utils.bugzilla_system.execute_bugzilla_call(
            call=mock_call, kwargs=kwargs
        )

        assert output == {"name": "foo"}

        mock_call.assert_called_with(product="Fedora", data={"component": "foo"})

    def test_execute_bugzilla_call_args_kwargs(self):
        """Assert that bugzilla call wrapper is working correctly with both args and kwargs."""
        mock_call = MagicMock()
        mock_call.return_value = {"name": "foo"}
        args = ("arg1",)
        kwargs = {"product": "Fedora", "data": {"component": "foo"}}

        output = toddlers.utils.bugzilla_system.execute_bugzilla_call(
            call=mock_call, args=args, kwargs=kwargs
        )

        assert output == {"name": "foo"}

        mock_call.assert_called_with(
            "arg1", product="Fedora", data={"component": "foo"}
        )

    @patch("time.sleep")
    def test_execute_bugzilla_call_exception(self, mock_time, caplog):
        """Assert that bugzilla call wrapper is working correctly when exception happens."""
        caplog.set_level(logging.DEBUG)
        mock_call = MagicMock()
        mock_call.side_effect = xmlrpc.client.Fault(50, "Error")
        kwargs = {"product": "Fedora", "data": {"component": "foo"}}

        with pytest.raises(xmlrpc.client.Fault):
            toddlers.utils.bugzilla_system.execute_bugzilla_call(
                call=mock_call, kwargs=kwargs, num_of_attempts=2
            )

        mock_call.assert_called_with(product="Fedora", data={"component": "foo"})

        assert caplog.records[0].getMessage() == "ERROR <Fault 50: 'Error'>"
        assert (
            caplog.records[1].getMessage()
            == "Query failed, going to try again in 10 seconds"
        )

        mock_time.assert_called_with(10)
