"""
Unit tests for `toddlers.utils.git`.
"""

from unittest.mock import MagicMock, Mock, patch

import pytest

from toddlers.utils.git import clone_repo, GitRepo


class TestCloneRepo:
    """
    Test class for `toddlers.utils.git.clone_repo` function.
    """

    @patch("toddlers.utils.git.git.Repo.clone_from")
    def test_clone_repo(self, clone_from_mock):
        """
        Assert that cloning works as expected.
        """
        remote = "https://example.com"
        destination = "/path/to/destination"

        repo_mock = Mock()
        clone_from_mock.return_value = repo_mock

        repo = clone_repo(remote, destination)

        clone_from_mock.assert_called_with(remote, destination)

        assert repo.repo == repo_mock


class TestGitRepoInit:
    """
    Test class for `toddlers.utils.git.GitRepo.__init__` method.
    """

    def test_init(self):
        """
        Assert that initialization works as expected.
        """
        repo_mock = Mock()

        repo = GitRepo(repo_mock)

        assert repo.repo == repo_mock


class TestGitRepoFirstCommit:
    """
    Test class for `toddlers.utils.git.GitRepo.first_commit` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = Mock()

        self.repo = GitRepo(repo_mock)

        assert self.repo.repo == repo_mock

    def test_first_commit(self):
        """
        Assert that first commit returning works fine.
        """
        branch = "branch"

        result = "hash1\nhash2"
        self.repo.repo.git.log.return_value = result

        commit = self.repo.first_commit(branch)

        self.repo.repo.git.log.assert_called_with(
            "--reverse", "--pretty=format:%H", branch
        )

        assert commit == "hash1"

    def test_first_commit_no_commit(self):
        """
        Assert that first commit returning works fine when
        no commit is returned.
        """
        branch = "branch"

        result = ""
        self.repo.repo.git.log.return_value = result

        commit = self.repo.first_commit(branch)

        self.repo.repo.git.log.assert_called_with(
            "--reverse", "--pretty=format:%H", branch
        )

        assert commit is None


class TestGitRepoDoesBranchExist:
    """
    Test class for `toddlers.utils.git.GitRepo.is_branch_exist` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = MagicMock()
        self.repo = GitRepo(repo_mock)
        assert self.repo.repo == repo_mock

    def test_does_branch_exist_not_exist(self):
        """
        Test that is_branch_exist returns False if branch does not exist.
        """
        branch_name = "non-existing-branch"
        self.repo.repo.remotes.origin.refs = []

        result = self.repo.does_branch_exist(branch_name)
        assert result is False, "Expected False, because branch is not exist"

    def test_does_branch_exist_exist(self):
        """
        Test that is_branch_exist returns True if branch is exist.
        """
        branch_name = "existing-branch"
        mock_ref = MagicMock()
        mock_ref.name = "origin/" + branch_name
        self.repo.repo.remotes.origin.name = "origin"
        self.repo.repo.remotes.origin.refs = [mock_ref]

        result = self.repo.does_branch_exist(branch_name)
        assert result is True, "Expected True, because branch is exist"


class TestGitRepoDoesBranchContainFile:
    """
    Test class for `toddlers.utils.git.GitRepo.is_branch_contains_file` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = MagicMock()
        self.repo = GitRepo(repo_mock)
        assert self.repo.repo == repo_mock

    def test_does_branch_contains_file_branch_not_exist(self):
        """
        Assert that is_branch_contains_file returns False if branch is not exist.
        """
        self.repo.repo.commit.side_effect = ValueError("Branch doesn't exists.")

        result = self.repo.does_branch_contains_file("non-existing-branch", "file_path")
        assert result is False, "Expected False, Because branch doesn't exists"

    def test_does_branch_contains_file_file_not_exist(self):
        """
        Assert that is_branch_contains_file returns False if file is not exist.
        """
        mock_commit = MagicMock()
        mock_blob = MagicMock()
        mock_blob.name = "some_file"
        mock_commit.tree = [mock_blob]
        self.repo.repo.commit.return_value = mock_commit

        result = self.repo.does_branch_contains_file("existing-branch", "my_file")
        assert result is False, "Expected False, because file doesn't exist"

    def test_does_branch_contains_file_file_exist(self):
        """
        Assert that is_branch_contains_file returns True if file is exist
        """
        mock_commit = MagicMock()
        mock_blob = MagicMock()
        mock_blob.name = "desired_file"
        mock_commit.tree = [mock_blob]
        self.repo.repo.commit.return_value = mock_commit

        result = self.repo.does_branch_contains_file("existing-branch", "desired_file")
        assert result is True, "Expected True, because file is exist"


class TestGitRepoGetLastCommitMessage:
    """
    Test class for `toddlers.utils.git.GitRepo.get_last_commit_message_of_specific_branch` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = MagicMock()

        self.repo = GitRepo(repo_mock)

        assert self.repo.repo == repo_mock

    def test_get_last_commit_message_error_was_raised(self):
        """
        Assert that commit message is none, when the branch name is wrong
        """
        branch = "wrong branch name"
        self.repo.repo.commit.side_effect = Exception(
            "{0} branch name is wrong".format(branch)
        )
        assert self.repo.get_last_commit_message(branch) is None

    def test_get_last_commit_message_issue_messages(self):
        """
        Assert that commit message will be correct on different types of branch name
        """
        mock_commit = MagicMock()
        mock_commit.message = "commit message"
        self.repo.repo.commit.return_value = mock_commit

        result_branch_without_prefix = self.repo.get_last_commit_message("branch")
        assert result_branch_without_prefix == "commit message"

        result_branch_with_prefix = self.repo.get_last_commit_message("origin/branch")
        assert result_branch_with_prefix == "commit message"

        result_rawhide_branch = self.repo.get_last_commit_message("rawhide")
        assert result_rawhide_branch == "commit message"


class TestGitRepoGetLastCommitDate:
    """
    Test class for `toddlers.utils.git.GitRepo.get_last_commit_date` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = Mock()

        self.repo = GitRepo(repo_mock)

        assert self.repo.repo == repo_mock

    def test_get_last_commit_date_error_raised(self):
        """
        Assert that error will be raised in case of wrong branch name
        """
        branch = "wrong branch name"
        self.repo.repo.commit.side_effect = Exception(branch + " branch name is wrong.")
        assert self.repo.get_last_commit_date(branch) is None

    def test_get_last_commit_date_different_type_of_branches(self):
        """
        Assert that getting last commit date works correctly on different branch names
        """
        mock_commit = MagicMock()
        committed_date = 71252
        mock_commit.committed_date = committed_date
        self.repo.repo.commit.return_value = mock_commit

        result_branch_without_prefix = self.repo.get_last_commit_date("branch")
        assert result_branch_without_prefix == committed_date

        result_branch_with_prefix = self.repo.get_last_commit_date("origin/branch")
        assert result_branch_with_prefix == committed_date

        result_rawhide_branch = self.repo.get_last_commit_date("rawhide")
        assert result_rawhide_branch == committed_date


class TestGitRepoRevertLastCommit:
    """
    Test class for `toddlers.utils.git.GitRepo.revert_last_commit` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = Mock()

        self.repo = GitRepo(repo_mock)

        assert self.repo.repo == repo_mock

    def test_revert_last_commit_success(self):
        """
        Assert that revert last commit process correctly.
        """
        mock_origin = MagicMock()
        mock_origin.url = "https://example.com"
        self.repo.repo.remote.return_value = mock_origin
        mock_git_cmd = MagicMock()
        self.repo.repo.git = mock_git_cmd

        self.repo.revert_last_commit("Revert message", "bot", "token", "feature_branch")
        self.repo.repo.git.checkout.assert_called_once_with("feature_branch")
        self.repo.repo.git.revert.assert_called_once_with("HEAD", no_edit=True)
        self.repo.repo.git.commit.assert_called_once_with(
            "--amend", "-m", "Revert message"
        )

        mock_git_cmd.push.assert_called_once_with(
            "-u", "https://bot:token@example.com", "feature_branch"
        )

    def test_revert_last_commit_revert_exception(self):
        mock_origin = MagicMock()
        self.repo.repo.remote.return_value = mock_origin
        self.repo.repo.git.revert.side_effect = Exception("Revert error")

        self.repo.revert_last_commit("Revert message", "bot", "token", "feature_branch")
        mock_origin.push.assert_not_called()


class TestGitRepoRetireBranch:
    """
    Test class for `toddlers.utils.git.GitRepo.retire_branch` method.
    """

    def setup_method(self):
        """
        Initialize the GitRepo object.
        """
        repo_mock = Mock()

        self.repo = GitRepo(repo_mock)

        assert self.repo.repo == repo_mock

    def test_retire_branch_success(self, tmp_path, monkeypatch):
        """
        Assert that retirement works correctly.
        """
        # Change working directory to temp directory
        # to prevent creating files in repo
        monkeypatch.chdir(tmp_path)
        mock_origin = MagicMock()
        mock_origin.url = "https://example.com"
        self.repo.repo.remote.return_value = mock_origin
        self.repo.repo.working_tree_dir = tmp_path
        mock_git_cmd = MagicMock()
        self.repo.repo.git = mock_git_cmd
        mock_commit = MagicMock()
        mock_blob = MagicMock()
        mock_blob.name = "README"
        mock_commit.tree = [mock_blob]
        self.repo.repo.commit.return_value = mock_commit

        self.repo.retire_branch("Retire branch", "bot", "token", "feature_branch")

        self.repo.repo.git.checkout.assert_called_once_with("feature_branch")
        self.repo.repo.index.remove.assert_called_once_with("README", working_tree=True)
        self.repo.repo.index.add.assert_called_once_with(["dead.package"])
        self.repo.repo.index.commit.assert_called_once_with("Retire branch")
        mock_git_cmd.push.assert_called_once_with(
            "-u", "https://bot:token@example.com", "feature_branch"
        )

    def test_retire_branch_exception(self):
        mock_origin = MagicMock()
        self.repo.repo.remote.return_value = mock_origin
        self.repo.repo.git.checkout.side_effect = Exception("Revert error")

        with pytest.raises(Exception):
            self.repo.retire_branch("Revert message", "bot", "token", "feature_branch")

        self.repo.repo.commit.assert_not_called()
