"""
Module for testing `toddlers.utils.package_summaries`.
"""

import bz2
import gzip
import hashlib
import logging
import lzma
import os
import tarfile
from unittest.mock import call, Mock

import pytest
import zstandard as zstd

from toddlers.utils.package_summaries import PackageSummaries


class TestGetPrimaryXML:
    """Test class for `toddlers.utils.package_summaries.PackageSummaries.get_primary_xml`."""

    def setup_method(self):
        """This will be run before every test case in this class."""
        self.ps = PackageSummaries()

    def test_get_primary_xml_no_response(self):
        """Assert that no response is handled correctly."""
        req = Mock()
        req.get.return_value = None
        self.ps.requests_session = req

        result = self.ps.get_primary_xml("/var/tmp/", "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result is None

    def test_get_primary_xml_parse_error(self, caplog):
        """Assert that response is handled correctly if it can't be parsed."""
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = ""
        req.get.return_value = resp_repomd
        self.ps.requests_session = req

        result = self.ps.get_primary_xml("/var/tmp/", "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result is None
        assert "Failed to parse https://foo.bar/repomd.xml" in caplog.text

    def test_get_primary_xml_no_data_nodes(self, caplog):
        """Assert that no data nodes found is handled correctly."""
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = "<repo></repo>"
        req.get.return_value = resp_repomd
        self.ps.requests_session = req

        with caplog.at_level(logging.DEBUG):
            result = self.ps.get_primary_xml("/var/tmp/", "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result is None
        assert "No primary.xml could be found in https://foo.bar" in caplog.text

    def test_get_primary_xml_multiple_primary_db(self, caplog):
        """Assert that multiple primary dbs found is handled correctly."""
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = """
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary"/>
          <repo:data type="primary"/>
        </root>
        """
        req.get.return_value = resp_repomd
        self.ps.requests_session = req

        with caplog.at_level(logging.DEBUG):
            result = self.ps.get_primary_xml("/var/tmp/", "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result is None
        assert (
            "More than one primary.xml could be found in https://foo.bar" in caplog.text
        )

    def test_get_primary_xml_missing_missing_location(self, caplog):
        """Assert that missing location node is handled correctly."""
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = """
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary"/>
        </root>
        """
        req.get.return_value = resp_repomd
        self.ps.requests_session = req

        with caplog.at_level(logging.DEBUG):
            result = self.ps.get_primary_xml("/var/tmp/", "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result is None
        assert (
            "No valid location found for primary.xml in https://foo.bar" in caplog.text
        )

    def test_get_primary_xml_missing_checksum_type(self, caplog):
        """Assert that missing location node is handled correctly."""
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = """
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="https://good.place"/>
          </repo:data>
        </root>
        """
        req.get.return_value = resp_repomd
        self.ps.requests_session = req

        with caplog.at_level(logging.DEBUG):
            result = self.ps.get_primary_xml("/var/tmp/", "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result is None
        assert (
            "No valid checksum information found for primary.xml in https://foo.bar"
            in caplog.text
        )

    def test_get_primary_xml_no_change(self, caplog, tmpdir):
        """Assert that nothing is done when there is no update of db."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")
        hashobj = getattr(hashlib, "sha1")()
        with open(filepath, "wb") as f:
            f.write(b"Hello, world!")
        with open(filepath, "rb") as f:
            hashobj.update(f.read())
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = f"""
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/db.gzip"/>
            <repo:open-checksum type="sha1">{hashobj.hexdigest()}</repo:open-checksum>
          </repo:data>
        </root>
        """
        req.get.return_value = resp_repomd
        self.ps.requests_session = req

        with caplog.at_level(logging.DEBUG):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        req.get.assert_called_with("https://foo.bar/repomd.xml", verify=True)

        assert result == filepath
        assert "No change of https://foo.bar/db.gzip" in caplog.text

    def test_get_primary_xml_no_local_file(self, caplog, tmpdir):
        """Assert that local db is downloaded when missing."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = """
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/db.xz"/>
            <repo:open-checksum type="sha1">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = lzma.compress(b"Hello, world!")
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call("https://foo.bar/db.xz", verify=True),
        ]

        assert result == filepath
        assert (
            "Downloading file: https://foo.bar/db.xz to {}/db.xz".format(tmpdir)
            in caplog.text
        )
        assert "Extracting {}/db.xz to {}".format(tmpdir, filepath) in caplog.text

        with open(filepath, "rb") as f:
            assert f.read() == b"Hello, world!"

    def test_get_primary_xml_different_hash(self, caplog, tmpdir):
        """Assert that local db is downloaded when missing."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")
        with open(filepath, "wb") as f:
            f.write(b"Hello, world!")
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = """
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/db.xz"/>
            <repo:open-checksum type="sha">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = lzma.compress(b"Hello, world!")
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call("https://foo.bar/db.xz", verify=True),
        ]

        assert result == filepath
        assert (
            "Downloading file: https://foo.bar/db.xz to {}/db.xz".format(tmpdir)
            in caplog.text
        )
        assert "Extracting {}/db.xz to {}".format(tmpdir, filepath) in caplog.text

        with open(filepath, "rb") as f:
            assert f.read() == b"Hello, world!"

    def test_get_primary_xml_tar_archive(self, caplog, tmpdir):
        """Assert that tar archived db is extracted."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")

        # Create a tar compressed test file
        test_file = os.path.join(tmpdir, "test")
        with open(test_file, "w") as f:
            f.write("Hello, world!")
        tar_file = os.path.join(tmpdir, "foo.tar")
        with tarfile.open(tar_file, "w:gz") as tar:
            tar.add(test_file, arcname=os.path.basename(test_file))

        archive_data = ""
        with open(tar_file, "rb") as f:
            archive_data = f.read()

        archive_name = "db.tar.gz"
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = f"""
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/{archive_name}"/>
            <repo:open-checksum type="sha1">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = archive_data
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call(f"https://foo.bar/{archive_name}", verify=True),
        ]

        assert result == filepath
        assert (
            "Downloading file: https://foo.bar/{0} to {1}/{0}".format(
                archive_name, tmpdir
            )
            in caplog.text
        )
        assert (
            "Extracting {}/{} to {}".format(tmpdir, archive_name, filepath)
            in caplog.text
        )

        with open(filepath, "r") as f:
            assert f.read() == "Hello, world!"

    def test_get_primary_xml_gzip_archive(self, caplog, tmpdir):
        """Assert that gzip archived db is extracted."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")

        archive_data = gzip.compress(b"Hello, world!")
        archive_name = "db.gz"
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = f"""
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/{archive_name}"/>
            <repo:open-checksum type="sha1">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = archive_data
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call(f"https://foo.bar/{archive_name}", verify=True),
        ]

        assert result == filepath
        assert (
            "Downloading file: https://foo.bar/{0} to {1}/{0}".format(
                archive_name, tmpdir
            )
            in caplog.text
        )
        assert (
            "Extracting {}/{} to {}".format(tmpdir, archive_name, filepath)
            in caplog.text
        )

        with open(filepath, "r") as f:
            assert f.read() == "Hello, world!"

    def test_get_primary_xml_bz2_archive(self, caplog, tmpdir):
        """Assert that bz2 archived db is extracted."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")

        archive_data = bz2.compress(b"Hello, world!")
        archive_name = "db.bz2"
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = f"""
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/{archive_name}"/>
            <repo:open-checksum type="sha1">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = archive_data
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call(f"https://foo.bar/{archive_name}", verify=True),
        ]

        assert result == filepath
        assert (
            "Downloading file: https://foo.bar/{0} to {1}/{0}".format(
                archive_name, tmpdir
            )
            in caplog.text
        )
        assert (
            "Extracting {}/{} to {}".format(tmpdir, archive_name, filepath)
            in caplog.text
        )

        with open(filepath, "r") as f:
            assert f.read() == "Hello, world!"

    def test_get_primary_xml_zstd_archive(self, caplog, tmpdir):
        """Assert that zstd archived db is extracted."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")

        cctx = zstd.ZstdCompressor()
        archive_data = cctx.compress(b"Hello, world!")
        archive_name = "db.zst"
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = f"""
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/{archive_name}"/>
            <repo:open-checksum type="sha1">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = archive_data
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            result = self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call(f"https://foo.bar/{archive_name}", verify=True),
        ]

        assert result == filepath
        assert (
            "Downloading file: https://foo.bar/{0} to {1}/{0}".format(
                archive_name, tmpdir
            )
            in caplog.text
        )
        assert (
            "Extracting {}/{} to {}".format(tmpdir, archive_name, filepath)
            in caplog.text
        )

        with open(filepath, "r") as f:
            assert f.read() == "Hello, world!"

    def test_get_primary_xml_unsupported_archive(self, caplog, tmpdir):
        """Assert that unsupported archived db is raise exception."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")

        archive_name = "db.zip"
        req = Mock()
        resp_repomd = Mock()
        resp_repomd.text = f"""
        <root xmlns:repo="http://linux.duke.edu/metadata/repo">
          <repo:data type="primary">
            <repo:location href="repodata/{archive_name}"/>
            <repo:open-checksum type="sha1">0</repo:open-checksum>
          </repo:data>
        </root>
        """
        resp_archive = Mock()
        resp_archive.content = b"Hello, world!"
        req.get.side_effect = [resp_repomd, resp_archive]
        self.ps.requests_session = req

        with caplog.at_level(logging.INFO):
            with pytest.raises(NotImplementedError):
                self.ps.get_primary_xml(tmpdir, "https://foo.bar")

        assert req.get.mock_calls == [
            call("https://foo.bar/repomd.xml", verify=True),
            call(f"https://foo.bar/{archive_name}", verify=True),
        ]

        assert (
            "Downloading file: https://foo.bar/{0} to {1}/{0}".format(
                archive_name, tmpdir
            )
            in caplog.text
        )
        assert (
            "Extracting {}/{} to {}".format(tmpdir, archive_name, filepath)
            in caplog.text
        )

        assert not os.path.exists(filepath)


class TestGetPackageSummaries:
    """Test class for `toddlers.utils.package_summaries.PackageSummaries.get_package_summaries`."""

    def setup_method(self):
        """This will be run before every test case in this class."""
        self.ps = PackageSummaries()

    def test_get_package_summaries(self, tmpdir):
        """Assert that summuries are correctly returned."""
        filepath = os.path.join(tmpdir, "distgit-bugzilla-sync-primary.xml")
        xml_data = """
        <metadata xmlns="http://linux.duke.edu/metadata/common" xmlns:rpm="http://linux.duke.edu/metadata/rpm" packages="54456">
          <package type="rpm">
            <name>foo</name>
            <summary>bar</summary>
          </package>
        </metadata>
        """  # noqa: E501
        with open(filepath, "w") as f:
            f.write(xml_data)

        config = {"temp_folder": tmpdir, "kojipkgs_url": "https://koji.pkgs.org"}
        mock_primary_xml = Mock()
        mock_primary_xml.return_value = filepath
        self.ps.get_primary_xml = mock_primary_xml

        result = self.ps.get_package_summaries(config)

        mock_primary_xml.assert_called_with(
            tmpdir, config["kojipkgs_url"] + "/repos/rawhide/latest/x86_64/repodata"
        )

        assert result == {"foo": "bar"}
