from collections import OrderedDict
from unittest.mock import ANY, MagicMock, patch

import pytest

from toddlers.utils.requests import (
    _Default,
    make_session,
    RETRY_DEFAULTS,
    TIMEOUT_DEFAULT,
    TimeoutHTTPAdapter,
)

TIMEOUTS_TO_TEST = (
    # default_timeout, timeout, expected_timeout
    (_Default, _Default, TIMEOUT_DEFAULT),
    (None, _Default, None),
    (_Default, 1, 1),
    (1, _Default, 1),
    (1, None, None),
    (1, 2, 2),
)

DEFAULT_TIMEOUTS_TO_TEST = {
    (dt, dt) if dt is not _Default else (dt, TIMEOUT_DEFAULT)
    for dt, t, e in TIMEOUTS_TO_TEST
}


class TestTimeoutHTTPAdapter:
    @pytest.mark.parametrize(
        "default_timeout, expected_timeout", DEFAULT_TIMEOUTS_TO_TEST
    )
    def test___init__(self, default_timeout, expected_timeout):
        adapter = TimeoutHTTPAdapter(timeout=default_timeout)
        assert adapter.timeout == expected_timeout

    @pytest.mark.parametrize(
        "default_timeout, timeout, expected_timeout", TIMEOUTS_TO_TEST
    )
    @patch("requests.adapters.HTTPAdapter.send")
    def test_send(self, send, default_timeout, timeout, expected_timeout):
        adapter = TimeoutHTTPAdapter(timeout=default_timeout)

        request = MagicMock()
        if timeout is _Default:
            adapter.send(request)
        else:
            adapter.send(request, timeout=timeout)

        send.assert_called_once_with(request, timeout=expected_timeout)


@pytest.mark.parametrize(
    "retry_conf, expected_total",
    (
        (None, RETRY_DEFAULTS["total"]),
        ({"total": RETRY_DEFAULTS["total"] + 1}, RETRY_DEFAULTS["total"] + 1),
    ),
)
@pytest.mark.parametrize("default_timeout, timeout, expected_timeout", TIMEOUTS_TO_TEST)
@patch("requests.sessions.extract_cookies_to_jar", new=MagicMock())
@patch("requests.adapters.HTTPAdapter.send")
def test_make_session(
    send, retry_conf, expected_total, default_timeout, timeout, expected_timeout
):
    resp = MagicMock()
    resp.url = "http://localhost"
    resp.is_redirect = False
    send.return_value = resp

    session = make_session(timeout=default_timeout, retry_conf=retry_conf)

    if timeout is _Default:
        session.get("http://localhost")
    else:
        session.get("http://localhost", timeout=timeout)

    send.assert_called_once_with(
        ANY,
        timeout=expected_timeout,
        verify=True,
        proxies=OrderedDict(),
        stream=False,
        cert=None,
    )

    for adapter in session.adapters.values():
        assert adapter.max_retries.total == expected_total
