"""
The playtime script is a way to trigger toddlers manually or at specific times.
It is meant to be run either manually or via a cron job simply by calling the
file with the toddler_name as first argument:

    python playtime.py <toddler_name>

This will result in a message being sent on the fedora-messaging message bus
with the topic: ``toddlers.trigger.<toddler_name``.
And **if** the toddler is set to react to these messages, it will trigger from
it.
"""

import argparse
import datetime
import logging
import sys

import fedora_messaging.api
import fedora_messaging.config
import fedora_messaging.exceptions


_log = logging.getLogger(__name__)


def setup_logging(log_level: int):
    handlers = []

    _log.setLevel(log_level)
    # We want all messages logged at level INFO or lower to be printed to stdout
    info_handler = logging.StreamHandler(stream=sys.stdout)
    handlers.append(info_handler)

    if log_level == logging.INFO:
        # In normal operation, don't decorate messages
        for handler in handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    logging.basicConfig(level=log_level, handlers=handlers)


def get_arguments(args):
    """Load and parse the CLI arguments."""
    parser = argparse.ArgumentParser(
        description="Sends a message on the fedora-messaging message bus to trigger a toddler "
        "(if that toddlers supports being triggered this way)."
    )
    parser.add_argument(
        "name",
        help="Name of the toddler to trigger",
    )
    parser.add_argument(
        "--conf",
        help="Fedora-messaging configuration file to use",
    )

    log_level_group = parser.add_mutually_exclusive_group()
    log_level_group.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.INFO,
        help="Enable debugging output",
    )

    return parser.parse_args(args)


def main(args=None):
    """Main function of the playtime script"""
    args = get_arguments(args)
    setup_logging(log_level=args.log_level)
    _log.debug("Log level set to: %s", args.log_level)

    if args.conf:
        _log.debug("Loading the fedora-messaging configuration at: %s", args.conf)
        fedora_messaging.config.conf.load_config(args.conf)

    try:
        _log.debug(
            "Send fedora-messaging trigger with topic: toddlers.trigger.%s", args.name
        )
        msg = fedora_messaging.api.Message(
            topic="toddlers.trigger.{}".format(args.name),
            body={"date": datetime.datetime.now(datetime.timezone.utc).isoformat()},
        )
        fedora_messaging.api.publish(msg)
        _log.debug("Message sent!")
    except fedora_messaging.exceptions.PublishReturned as err:
        _log.exception(f"Fedora Messaging broker rejected message {msg.id}: {err}")
        return 1
    except fedora_messaging.exceptions.ConnectionException as err:
        _log.exception(f"Error sending message {msg.id}: {err}")
        return 1
    except Exception as err:
        _log.exception(f"Error sending fedora-messaging message: {err}")
        return 1


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())
