"""
This script takes as input the fedmsg messages published under the topic
``toddlers.trigger.check_commit_rights`` and searches for any user not
member of the packager group which still have commit rights.

Authors:    Mattia Verga <mattia.verga@proton.me>

"""

import argparse
import logging
import sys

import tomllib

from toddlers.base import ToddlerBase
from toddlers.utils import fedora_account
from toddlers.utils import notify
from toddlers.utils.misc import merge_dicts
from toddlers.utils.requests import make_session


_log = logging.getLogger(__name__)


class CheckCommitRights(ToddlerBase):
    """Listens to messages sent by playtime (which lives in toddlers) and searches
    for any user not member of the packager group which still have commit rights.
    """

    name = "check_commit_rights"

    amqp_topics = [
        "org.fedoraproject.*.toddlers.trigger.check_commit_rights",
    ]

    def __init__(self):
        super().__init__()
        self.requests_session = make_session()

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "toddlers.trigger.check_commit_rights"
        )

    def process(self, config, message, send_email=True):
        """Looks for users with commit rights not in packager group."""
        exclude_users = config.get("exclude_users", None)
        if exclude_users is None:
            raise Exception("Invalid toddler configuration, no `exclude_users` defined")

        notify_emails = config.get("notify_emails")
        if not notify_emails and send_email:
            raise Exception("Invalid toddler configuration, no `notify_emails` defined")

        _log.info("### Setting up connection to FAS ###")
        fedora_account.set_fasjson(config)

        _log.info("### Getting packagers list ###")
        packagers_list = fedora_account.get_group_member("packager")
        _log.info(f"Found {len(packagers_list)} packagers")
        if not packagers_list:
            raise Exception("Something wrong occurred, I found no packagers. Exiting.")

        _log.info("### Looking for users with commit rights ###")
        page = 1
        proj_num = 0
        not_packagers = []
        while True:
            req = self.requests_session.get(
                f"{config['dist_git_url']}/api/0/projects?fork=false&namespace=rpms"
                f"&per_page=100&page={page}"
            )
            data = req.json()
            projects = data.get("projects", [])
            total_pages = data.get("pagination", dict()).get("pages", 1)
            for proj in projects:
                proj_num += 1
                users = proj.get("access_users", dict())
                for user in (
                    users.get("owner")
                    + users.get("admin")
                    + users.get("collaborator")
                    + users.get("commit")
                ):
                    if user in exclude_users:
                        continue
                    if user not in packagers_list:
                        not_packagers.append(user)
            _log.info(f"Done {page} pages out of {total_pages}.")
            if page >= total_pages:
                break
            page += 1

        not_packagers_set = sorted(set(not_packagers))
        _log.info(
            f"### Found {len(not_packagers_set)} users with "
            "commit privileges but not in packager group ###"
        )
        for u in not_packagers_set:
            _log.info(u)

        if len(not_packagers_set) and send_email:
            _log.info("### Sending email to admins ###")
            users_list = "\n".join([f"- {username}" for username in not_packagers_set])
            message = f"""Dear Admin,

The periodic check on user commit rights on src.fp.o has identified {len(not_packagers_set)}
users which aren't member of the packager group. The list of them is here:

{users_list}

Please check is these rights are correct, they can be a leftover from when the users
were removed, or removed themselves, from the packager group.

Have a wonderful day and see you (maybe?) at the next run!

"""
            notify.send_email(
                to_addresses=config["notify_emails"],
                from_address=config["admin_email"],
                subject="Ex-packagers still having commit rights",
                content=message,
                mail_server=config["mail_server"],
            )


# The following code allows to run this script stand-alone if needed.


def setup_logging(log_level: int):
    handlers = []

    _log.setLevel(log_level)
    # We want all messages logged at level INFO or lower to be printed to stdout
    info_handler = logging.StreamHandler(stream=sys.stdout)
    handlers.append(info_handler)

    if log_level == logging.INFO:
        # In normal operation, don't decorate messages
        for handler in handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    logging.basicConfig(level=log_level, handlers=handlers)


def get_arguments(args):
    """Load and parse the CLI arguments."""
    parser = argparse.ArgumentParser(
        description="Looks for users with commit rights not in packager group"
    )
    parser.add_argument(
        "conf",
        help="Configuration file",
    )

    log_level_group = parser.add_mutually_exclusive_group()
    log_level_group.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        dest="log_level",
        const=logging.WARNING,
        default=logging.INFO,
        help="Be less talkative",
    )
    log_level_group.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        help="Enable debugging output",
    )

    return parser.parse_args(args)


def main(args):
    """Schedule the first test and run the scheduler."""
    args = get_arguments(args)
    setup_logging(log_level=args.log_level)

    with open(args.conf, "rb") as conf_fp:
        base_config = tomllib.load(conf_fp)

    default_config = base_config.get("consumer_config", {}).get("default", {})
    private_config = base_config.get("consumer_config", {}).get(
        "check_commit_rights", {}
    )
    config = merge_dicts(default_config, private_config)

    CheckCommitRights().process(
        config=config,
        message={},
        send_email=False,
    )


if __name__ == "__main__":  # pragma: no cover
    try:
        main(sys.argv[1:])
    except KeyboardInterrupt:
        pass
