"""
This script takes as input the fedmsg messages published under the topic
``toddlers.trigger.packager_bugzilla_sync`` and runs a sync of all packagers
found in FAS to bugzilla so that they can do things like editing flags on Fedora
tickets.

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import argparse
import logging
import sys

import tomllib

from toddlers.base import ToddlerBase
from toddlers.utils import bugzilla_system
from toddlers.utils import fedora_account
from toddlers.utils import notify
from toddlers.utils.misc import merge_dicts
from toddlers.utils.requests import make_session


_log = logging.getLogger(__name__)


class CheckEmailOverrides(ToddlerBase):
    """Listens to messages sent by playtime (which lives in toddlers) and check
    if the `email_overrides.toml` file stored in ansible is up to date.
    """

    name = "check_email_overrides"

    amqp_topics = [
        "org.fedoraproject.*.toddlers.trigger.check_email_overrides",
    ]

    def __init__(self):
        super().__init__()
        self.requests_session = make_session()

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "toddlers.trigger.check_email_overrides"
        )

    def process(self, config, message, dry_run=False):
        """Process a given message."""

        url = config.get("email_overrides_url")
        if not url:
            raise Exception(
                "Invalid toddler configuration, no `email_overrides_url` defined"
            )

        _log.info("Loading the email_overrides.toml file")
        req = self.requests_session.get(url)
        if not req.ok:
            raise Exception(
                "Could not retrieve the email_overrides.toml file at: %s - returned %s",
                req.url,
                req.status_code,
            )

        data = tomllib.loads(req.text)

        _log.info("Setting up connection to FAS")
        fedora_account.set_fasjson(config)

        _log.info("Setting up connection to bugzilla")
        bugzilla_system.set_bz(config)

        logs = []
        for fas_email, bz_email in data.items():
            _log.debug("Processing mapping %s (fas) -> %s (bz)", fas_email, bz_email)
            fas_user = fedora_account.get_user_by_email(fas_email)
            if not fas_user:
                logs.append(
                    f"The mapping {fas_email} (fas) -> {bz_email} (bz) did not "
                    "map to a FAS account"
                )
            bz_user = bugzilla_system.get_user(bz_email)
            if not bz_user:
                logs.append(
                    f"The mapping {fas_email} (fas) -> {bz_email} (bz) did not "
                    "map to a bugzilla account"
                )

        _log.info(f"{len(logs)} invalid mapping found")
        if logs:
            _log.info("- %s", "\n- ".join(logs))
            logs_text = "\n- ".join(logs)
            notify.notify_admins_on_invalid_entry_in_email_override(
                to_addresses=[config["admin_email"]],
                from_address=config["admin_email"],
                mail_server=config["mail_server"],
                url=url,
                logs_text=logs_text,
            )


# We have had the situation in the past where we've had to check a specific
# account, so the following code allows to run this script stand-alone if
# needed.


def get_arguments(args):
    """Load and parse the CLI arguments."""
    parser = argparse.ArgumentParser(
        description="Check the state of the email_overrides file"
    )
    parser.add_argument(
        "conf",
        help="Configuration file",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Do not open a ticket in pagure",
    )

    log_level_group = parser.add_mutually_exclusive_group()
    log_level_group.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        dest="log_level",
        const=logging.WARNING,
        default=logging.INFO,
        help="Be less talkative",
    )
    log_level_group.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        help="Enable debugging output",
    )

    return parser.parse_args(args)


def setup_logging(log_level: int):
    handlers = []

    _log.setLevel(log_level)
    # We want all messages logged at level INFO or lower to be printed to stdout
    info_handler = logging.StreamHandler(stream=sys.stdout)
    handlers.append(info_handler)

    if log_level == logging.INFO:
        # In normal operation, don't decorate messages
        for handler in handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    logging.basicConfig(level=log_level, handlers=handlers)


def main(args):
    """Schedule the first test and run the scheduler."""
    args = get_arguments(args)
    setup_logging(log_level=args.log_level)

    with open(args.conf, "rb") as conf_fh:
        base_config = tomllib.load(conf_fh)

    default_config = base_config.get("consumer_config", {}).get("default", {})
    private_config = base_config.get("consumer_config", {}).get(
        "check_email_overrides", {}
    )
    config = merge_dicts(default_config, private_config)

    CheckEmailOverrides().process(
        config=config,
        message={},
        dry_run=args.dry_run,
    )


if __name__ == "__main__":  # pragma: no cover
    try:
        main(sys.argv[1:])
    except KeyboardInterrupt:
        pass
