"""
The script is triggered by fedora-messaging messages published under the topic
``ipa.group_remove_member.v1`` and checks if a user was removed
from group "packagers". If yes, user is removed from packager related groups
as well.
"""

import logging
import os

import ipalib

from toddlers.base import ToddlerBase
from toddlers.exceptions import PagureError
from toddlers.utils import pagure

_log = logging.getLogger(__name__)


class CleanPackagerGroups(ToddlerBase):
    """Listens to messages to check for a membership removal message from FAS"""

    name: str = "clean_packagers_groups"
    amqp_topics: list[str] = [
        "org.fedoraproject.*.ipa.group_remove_member.v1",
    ]
    pagure: pagure.Pagure

    def __init__(self):
        super().__init__()
        self.pagure_url = None
        self.dist_git = None
        self._ipa_session = None

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "ipa.group_remove_member.v1"
        )

    def _get_ipa_session(self, config):  # pragma: no cover
        """Makes an ipa session that handles logging in"""
        if self._ipa_session is None:
            conf_vars = ("KRB5_CONFIG", "IPA_CONFDIR")
            for conf_var in conf_vars:
                if not config.get(conf_var):
                    continue
                os.environ[conf_var] = config[conf_var]
            self._ipa_session = ipalib.api
            self._ipa_session.bootstrap(context="custom")
            self._ipa_session.load_plugins()
            self._ipa_session.finalize()
            self._ipa_session.Backend.rpcclient.connect()
        return self._ipa_session

    def process(self, config, message):
        """Process a given message. Remove user that is being removed from
        a packager group from all groups in distgit"""
        group = message.group
        watched_groups = config.get("watched_groups")

        # If user is not removed from packager group, bail
        if group not in watched_groups:
            _log.info(
                f"User(s) {', '.join(message.user_names)} was not removed from packager group, "
                f"but from {group} group, nothing to do."
            )
            return

        self.dist_git = pagure.set_pagure(
            {
                "pagure_url": config.get("dist_git_url"),
                "pagure_api_key": config.get("dist_git_token"),
            }
        )
        _log.info("Fetching all distgit groups:")
        distgit_groups = self.dist_git.get_all_groups()
        _log.info(f"Distgit groups found: {distgit_groups}")

        if not distgit_groups:
            _log.info("No distgit groups found, bailing.")
            return

        ipa_session = self._get_ipa_session(config)

        for user in message.user_names:
            self._process_removal(config, user, group, distgit_groups, ipa_session)

    def _process_removal(self, config, user, group, distgit_groups, ipa_session):
        _log.info(
            f"User {user} was removed from {group} group, removing from "
            "packager-related groups as well."
        )

        # Remove user from the distgit group that triggered the toddler
        try:
            self.dist_git.remove_member_from_group(user, group)
            _log.info(f"User {user} removed from distgit group: {group}")
        except PagureError:
            _log.exception(
                f"Error while removing user {user} from distgit group {group}"
            )

        _log.info(f"Fetching groups user {user} is member of in ipa:")
        ipa_user = ipa_session.Command.user_show(uid=user)
        ipa_groups = ipa_user["result"]["memberof_group"]
        _log.info(f"Ipa groups: {ipa_groups}")

        intersection = list(set(distgit_groups) & set(ipa_groups))

        if not intersection:
            _log.info(
                f"Seems that user {user} is not member of any "
                "groups shared between ipa and distgit"
            )
            return
        _log.info(
            f"User {user} should be removed from following groups: {intersection}"
        )

        for group in intersection:
            try:
                # Remove user from ipa groups:
                ipa_session.Command.group_remove_member(cn=group, user=user)
                _log.info(f"User {user} removed from ipa group: {group}")
            except Exception:
                _log.exception(
                    f"Error while removing user {user} from ipa group {group}"
                )
            try:
                # Remove user from distgit groups:
                # pagure token with acls=group_modify needed
                self.dist_git.remove_member_from_group(user, group)
                _log.info(f"User {user} removed from distgit group: {group}")
            except PagureError:
                _log.exception(
                    f"Error while removing user {user} from distgit group {group}"
                )
