"""
This script takes as input the fedora messages published under the topic
``toddlers.trigger.distgit_bugzilla_sync`` and runs a sync of package ownership
from distgit to bugzilla.

Authors:    Michal Konecny <mkonecny@redhat.com>

"""

import argparse
import collections
import logging
from operator import itemgetter
import re
import sys
import time
from typing import Any, cast, Dict, Iterator, Mapping, Optional
import xmlrpc

import tomllib

try:
    import tqdm
except ImportError:
    tqdm = None

from ..base import ToddlerBase
from ..utils import bodhi, bugzilla_system, fedora_account, notify, pagure
from ..utils.package_summaries import PackageSummaries
from ..utils.requests import make_session

_log = logging.getLogger(__name__)


class DistgitBugzillaSync(ToddlerBase):
    """Listens to messages sent by playtime (which lives in toddlers) to sync
    package ownership from distgit to bugzilla.
    """

    name = "distgit_bugzilla_sync"
    amqp_topics = ["org.fedoraproject.*.toddlers.trigger.distgit_bugzilla_sync"]

    def __init__(self):
        super().__init__()
        self.requests_session = make_session()
        # List of distgit project
        self.pagure_projects = []
        # Dictionary of errors to send notifications about
        self.errors = collections.defaultdict(list)
        # Mapping of namespace to product, will be filled from configuration
        self.namespace_to_product = {}
        # Mapping of product to branch regex, will be filled from configuration
        self.product_to_branch_regex = {}

        self.dist_git = None
        self.bodhi = None

    def accepts_topic(self, topic: str) -> bool:
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.

        :arg topic: Topic to check.

        :returns: True if topic is accepted, False otherwise.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "toddlers.trigger.distgit_bugzilla_sync"
        )

    def process(
        self,
        config: dict,
        message: dict,
        projects: Iterator[str] = cast(Iterator[str], []),
        print_fas_names: bool = False,
        print_no_change: bool = False,
        dry_run: bool = False,
    ) -> None:
        """Process a given message.

        :arg config: Toddlers configuration
        :arg message: Message to process
        :arg projects: Distgit projects to sync
        :arg print_fas_names: Print FAS names instead of
            bugzilla usernames (e-mails)
        :arg print_no_change: Log message even for projects
            where no change was done
        :arg dry_run: Don't do any change in bugzilla
        """
        self.dist_git = pagure.set_pagure(
            {
                "pagure_url": config.get("dist_git_url"),
                "pagure_api_key": config.get("dist_git_token"),
            }
        )

        self.bodhi = bodhi.set_bodhi(config)

        try:
            with open(config["email_overrides_file"], "rb") as overrides_fh:
                email_overrides = tomllib.load(overrides_fh)
        except Exception:
            _log.error("Failed to load the file containing the email-overrides")
            raise

        # Fill the class attributes we will need later
        for product, values in config["products"].items():
            if "namespace" in values:
                if values["namespace"] in self.namespace_to_product:
                    self.namespace_to_product[values["namespace"]].append(product)
                else:
                    self.namespace_to_product[values["namespace"]] = [product]

        self.product_to_branch_regex = {
            n: re.compile(p["branch_regex"])
            for n, p in config["products"].items()
            if "branch_regex" in p
        }
        times = {"start": time.time()}

        # Clean the errors
        self.errors = collections.defaultdict(list)

        _log.debug("Building a cache of the rpm package summaries")
        rpm_summaries = PackageSummaries().get_package_summaries(config)

        self._get_pagure_projects(config["dist_git_url"], project_list=projects)

        self._add_branches_products_and_summaries(
            rpm_summaries,
            config["ignorable_accounts"],
        )

        times["data structure end"] = time.time()

        _log.info("Setting up connection to FAS")
        fedora_account.set_fasjson(config)

        _log.info("Build bugzilla mail cache for users")
        bz_mail_to_fas = {}
        fas_to_bz_mail = {}

        for project in self.pagure_projects:
            if "products_poc" in project:
                for product, product_poc in project["products_poc"].items():
                    username = product_poc
                    if username and username not in fas_to_bz_mail:
                        bz_mail = self._get_bz_mail(username, email_overrides)
                        if bz_mail:
                            fas_to_bz_mail[username] = bz_mail
                            bz_mail_to_fas[bz_mail] = username
            for watcher in project["watchers"]:
                if watcher not in fas_to_bz_mail:
                    bz_mail = self._get_bz_mail(watcher, email_overrides)
                    if bz_mail:
                        fas_to_bz_mail[watcher] = bz_mail
                        bz_mail_to_fas[bz_mail] = watcher

        times["FAS cache building end"] = time.time()

        _log.info("Setting up connection to Bugzilla")
        bugzilla_system.set_bz(config)

        _log.debug("Building bugzilla's products in-memory cache")
        product_cache = cast(Dict[str, Any], {})
        for collection in config["products"]:
            product_cache[collection] = bugzilla_system.get_product_info_packages(
                collection
            )
        times["BZ cache building end"] = time.time()

        if dry_run:
            _log.info("Querying bugzilla but not doing anything")
        else:
            _log.info("Updating bugzilla")

        # If the import fails, no progress bar
        # At DEBUG or below, we're showing things at each iteration so the progress
        # bar doesn't look good.
        # At WARNING or above, we do not want to show anything.
        if (
            tqdm is not None and _log.getEffectiveLevel() == logging.INFO
        ):  # pragma no cover
            self.pagure_projects = tqdm.tqdm(self.pagure_projects)

        for project in sorted(self.pagure_projects, key=itemgetter("name")):
            for product in project["products"]:
                # Turn the watchers list into something usable by bugzilla
                cc_list = []
                cc_list_lower = []
                for watcher in project["watchers"]:
                    if watcher in fas_to_bz_mail:
                        cc_list.append(fas_to_bz_mail[watcher])
                        cc_list_lower.append(fas_to_bz_mail[watcher].lower())
                    else:
                        self.errors["mails"].append(
                            f"`{watcher}` has no bugzilla_email or mailing_list set on "
                            f"`{product}/{project['name']}`"
                        )

                # Add owner to the cclist so comaintainers taking over a bug don't
                # have to do this manually
                owner = project["products_poc"][product]
                if owner in fas_to_bz_mail:
                    if fas_to_bz_mail[owner].lower() not in cc_list_lower:
                        cc_list.append(fas_to_bz_mail[owner])
                else:
                    # The owner doesn't have a bugzilla e-mail on FAS account
                    # Not much we can do about it, log the error and continue
                    self.errors["mails"].append(
                        f"`{owner}` has no bugzilla_email or mailing_list set on "
                        f"`{product}/{project['name']}`"
                    )
                    continue

                qa_contact_email = config.get("default_qa_contact") or None

                package = project["name"]
                try:
                    if package in product_cache.get(product, []):
                        bugzilla_system.edit_component(
                            owner=fas_to_bz_mail[owner],
                            product=product,
                            package=project["name"],
                            component=product_cache[product][package],
                            cc_list=cc_list,
                            versions=config["products"][product]["versions"],
                            description=project["summary"],
                            fas_users_info=bz_mail_to_fas,
                            retired=project["products_retired"][product],
                            print_fas_names=print_fas_names,
                            print_no_change=print_no_change,
                            dry_run=dry_run,
                        )
                    else:
                        bugzilla_system.add_component(
                            product=product,
                            owner=fas_to_bz_mail[owner],
                            package=project["name"],
                            qa_contact=qa_contact_email,
                            cc_list=cc_list,
                            fas_users_info=bz_mail_to_fas,
                            description=project["summary"],
                            retired=project["products_retired"][product],
                            print_fas_names=print_fas_names,
                            dry_run=dry_run,
                        )
                except xmlrpc.client.ProtocolError as e:
                    # Unrecoverable and likely means that nothing is going to
                    # succeed.
                    self.errors["bugzilla_raw"].append(str(e.args))
                    self.errors["bugzilla"].append(
                        f"Failed to update: `{product}/{project['name']}`:"
                        f"\n    {e}"
                        f"\n    {e.args}"
                    )
                    break
                except xmlrpc.client.Error as e:
                    # An error occurred in the xmlrpc call. Shouldn't happen but
                    # we better see what it is
                    self.errors["bugzilla_raw"].append(
                        "%s -- %s" % (project["name"], e.args[-1])
                    )
                    self.errors["bugzilla"].append(
                        f"Failed to update: `{product}/{project['name']}`:"
                        f"\n    {e}"
                        f"\n    {e.args}"
                    )

        if self.errors:
            mail_server = config["mail_server"]
            admin_email = config["admin_email"]
            notify_admins = config["notify_admins"]

            # Build the report for the admins
            report = ["ERROR REPORT"]
            final_keys = []
            for key in ["configuration", "mails", "bugzilla"]:
                if self.errors[key]:
                    final_keys.append(key)
                    report.append(key)
                    report.append("  - {}".format("\n  - ".join(self.errors[key])))
                    report.append("")

            if dry_run:
                _log.debug("*" * 80)
                _log.debug("\n".join(report))

            # Do not send the email in dryrun or when the error only relates to
            # configuration (which will always happen for flatpaks and tests)
            _log.debug("email error key: %s", final_keys)
            if not dry_run and final_keys != ["configuration"]:
                notify.notify_admins_distgit_sync_error(
                    mail_server, admin_email, notify_admins, report
                )

        times["end"] = time.time()

        _log.info("   ----------")
        _log.info("Building the data structure")
        delta = times["data structure end"] - times["start"]
        _log.info("  Ran on %2f seconds -- ie: %2f minutes", delta, delta / 60)

        _log.info("Building the FAS cache")
        delta = times["FAS cache building end"] - times["data structure end"]
        _log.info("  Ran on %2f seconds -- ie: %2f minutes", delta, delta / 60)

        _log.info("Building the bugzilla cache")
        delta = times["BZ cache building end"] - times["FAS cache building end"]
        _log.info("  Ran on %2f seconds -- ie: %2f minutes", delta, delta / 60)

        _log.info("Interacting with bugzilla")
        delta = times["end"] - times["BZ cache building end"]
        _log.info("  Ran on %2f seconds -- ie: %2f minutes", delta, delta / 60)

        _log.info("Total")
        delta = times["end"] - times["start"]
        _log.info("  Ran on %2f seconds -- ie: %2f minutes", delta, delta / 60)

    def _get_pagure_projects(
        self, dist_git_url: str, project_list: Iterator[str] = cast(Iterator[str], [])
    ) -> None:
        """Builds a list of the projects on pagure in `self.pagure_projects`.
        Where every item is a dict containing:
        - the namespace of the project
        - the name of the project
        - the point of contact of this project (ie: the default assignee
          in bugzilla)
        - the watchers of this project (ie: the initial CC list in bugzilla)

        :arg dist_git_url: Pagure distgit URL
        :arg project_list: List of the projects that will be in the returned list.
            If not given, returns every project in pagure. Default set to `None`.
        """
        # Get the initial ownership and CC data from pagure
        # This part is easy.
        poc_url = dist_git_url + "/extras/pagure_poc.json"
        _log.debug("Querying %r for points of contact.", poc_url)
        pagure_namespace_to_poc = self.requests_session.get(poc_url, timeout=120).json()

        cc_url = dist_git_url + "/extras/pagure_bz.json"
        _log.debug("Querying %r for initial cc list.", cc_url)
        pagure_namespace_to_cc = self.requests_session.get(cc_url, timeout=120).json()

        # Combine and collapse those two into a single list:
        self.pagure_projects = []
        project_set = {tuple(p.split("/", 1)) for p in project_list}
        for namespace, entries in pagure_namespace_to_poc.items():
            for name, poc in entries.items():
                if not project_set or (namespace, name) in project_set:
                    if namespace in self.namespace_to_product:
                        self.pagure_projects.append(
                            {
                                "namespace": namespace,
                                "name": name,
                                "poc": poc["fedora"],
                                "epelpoc": poc["epel"],
                                "watchers": pagure_namespace_to_cc[namespace][name],
                            }
                        )

    def _add_branches_products_and_summaries(
        self,
        rpm_summaries: Mapping[str, str],
        ignorable_accounts: list,
    ) -> None:
        """For each project retrieved, this method adds branches, products
        and summary information.
        The products are determined based on the branches.
        The summaries are coming from the primary.xml file of the Rawhide repodata
        in Koji.

        :arg rpm_summaries: Package summaries from koji
        :arg ignorable_accounts: Accounts to ignore when filling up watchers list.
        """
        for idx, project in enumerate(self.pagure_projects):
            # Summary
            summary = None
            if project["namespace"] == "rpms":
                summary = rpm_summaries.get(project["name"])
            project["summary"] = summary

            try:
                project["branches"] = self.dist_git.get_branches(
                    project["namespace"], project["name"]
                )
            except pagure.PagureError:
                _log.exception(
                    "Could not retrieve the git branches for %s/%s",
                    project["namespace"],
                    project["name"],
                )
                project["branches"] = []

            if not project["branches"]:
                self.errors["dist-git"].append(
                    f"No branch found for {project['namespace']}/{project['name']}"
                )

            # Products
            products = set()
            # Determine product for project
            # First based on namespace
            namespace = project.get("namespace")
            if len(self.namespace_to_product[namespace]) > 1:
                # When we have more than one entry for the namespace, use branch regex
                for product in self.namespace_to_product[namespace]:
                    for branch in project.get("branches", []):
                        regex = self.product_to_branch_regex[product]
                        if regex.match(branch):
                            products.add(product)
                            break
            else:
                products.add(self.namespace_to_product[project["namespace"]][0])
            project["products"] = list(sorted(products))

            active_branches = self.bodhi.get_active_branches()
            products_poc = {}
            products_retired = {}
            for product in products:
                owner = project["poc"]
                # Check if the project is retired in dist_git, and if so set assignee to orphan.
                if self._is_retired(product, project, active_branches):
                    owner = "orphan"
                    products_retired[product] = True
                else:
                    products_retired[product] = False

                # Check if the Bugzilla ticket assignee has been manually overridden
                if product == "Fedora EPEL":
                    products_poc[product] = project["epelpoc"]
                else:
                    products_poc[product] = owner

            project["products_poc"] = products_poc
            project["products_retired"] = products_retired

            # Clean up the watchers we never want to sync to bugzilla
            # If these users are POC for a project, things will not work, which
            # is expected/desired.
            for user in ignorable_accounts or []:
                if user in (project.get("watchers") or []):
                    project["watchers"].remove(user)

            self.pagure_projects[idx] = project

    def _get_bz_mail(
        self, username: str, email_overrides: Mapping[str, str]
    ) -> Optional[str]:
        """Get bugzilla e-mail for username.

        :arg username: FAS username
        :arg email_overrides: E-mail overrides for bugzilla

        :returns: Bugzilla e-mail
        """
        if username.startswith("@"):
            return fedora_account.get_bz_email_group(username[1:], email_overrides)
        return fedora_account.get_bz_email_user(username, email_overrides)

    def _is_retired(self, product: str, project: dict, active_branches: list) -> bool:
        """Check if the project is retired.
        The project is considered retired if there is no active branch.

        :arg product: Product to determine branches for which to look (example: "Fedora")
        :arg project: Project to check
        :arg active_branches: Active branches in Bodhi

        :returns: Retired status.
        """
        branches = project["branches"]
        branch_regex = self.product_to_branch_regex.get(product)
        for branch in branches:
            # Only check the branches that match the regex for Product in bugzilla
            if branch_regex.match(branch) and branch in active_branches:
                # One branch matching the product regex is still active
                # The project is not retired yet
                if not self.dist_git.is_retired_on_branch(project["name"], branch):
                    return False
        # No branch matching the regex is active, the project is retired in the product
        return True


# In case this code needs to be run manually for some projects
# the following code allows to run this script stand-alone if
# needed.


def _get_arguments(args):
    """Load and parse the CLI arguments.

    :arg args: Script arguments

    :returns: Parsed arguments
    """
    parser = argparse.ArgumentParser(
        description="Sync distgit package info with bugzilla"
    )

    parser.add_argument(
        "conf",
        help="Configuration file",
    )

    parser.add_argument(
        "-p",
        "--project",
        dest="projects",
        nargs="+",
        help="Update one or more projects (provided as namespace/name), "
        "in all of its products",
    )

    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
        help="Do not change anything on bugzilla",
    )

    parser.add_argument(
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.INFO,
        help="Enable debugging output",
    )

    parser.add_argument(
        "--print-fas-names",
        action="store_true",
        default=False,
        help="Print FAS names rather than email addresses in output, useful when pasting into "
        "public fora",
    )

    parser.add_argument(
        "--print-no-change",
        action="store_true",
        default=False,
        help="Print elements that are not being changed as they are checked",
    )

    return parser.parse_args(args)


def _setup_logging(log_level: int) -> None:
    """Setup the logging level.

    :arg log_level: Log level to set
    """
    handlers = []

    _log.setLevel(log_level)
    # We want all messages logged at level INFO or lower to be printed to stdout
    info_handler = logging.StreamHandler(stream=sys.stdout)
    handlers.append(info_handler)

    if log_level == logging.INFO:
        # In normal operation, don't decorate messages
        for handler in handlers:
            handler.setFormatter(logging.Formatter("%(message)s"))

    logging.basicConfig(level=log_level, handlers=handlers)


def main(args):
    """Main function."""
    args = _get_arguments(args)
    _setup_logging(log_level=args.log_level)

    with open(args.conf, "rb") as conf_fh:
        config = tomllib.load(conf_fh)
    parsed_config = config.get("consumer_config", {}).get("default", {})
    parsed_config.update(
        config.get("consumer_config", {}).get("distgit_bugzilla_sync", ())
    )
    DistgitBugzillaSync().process(
        config=parsed_config,
        message={},
        projects=args.projects,
        print_fas_names=args.print_fas_names,
        print_no_change=args.print_no_change,
        dry_run=args.dry_run,
    )


if __name__ == "__main__":  # pragma: no cover
    try:
        main(sys.argv[1:])
    except KeyboardInterrupt:
        pass
