"""
This plugin takes as input the Fedora messages published under the topic
``pagure.git.receive`` (for dist-git) and sends out emails to scm-commits-list.
"""

import json
import logging
from typing import Any, Optional

from fedora_messaging.api import Message
from fedora_messaging_git_hook_messages import CommitV1

from toddlers.base import ToddlerBase
from toddlers.exceptions import ConfigError
from toddlers.utils.notify import send_email

log = logging.getLogger(__name__)


class DistGitCommitProcessor(ToddlerBase):
    """A processor which sends emails about commits to dist-git repositories to
    scm-commits-list.
    """

    name: str = "distgit_commit_processor"

    amqp_topics: list[str] = ["org.fedoraproject.*.git.receive"]

    def accepts_topic(self, topic: str) -> bool:
        """Check whether this toddler is interested in messages of a topic.

        :arg topic: Topic to check.

        :returns: Whether or not this toddler accepts the topic.
        """
        return (
            topic.startswith("org.fedoraproject.")
            and topic.endswith(".git.receive")
            and not topic.endswith(".pagure.git.receive")  # different type of message
        )

    def ignore_commit(self, message: CommitV1) -> bool:
        """Determine if a message pertains to a forked repo.

        :arg message: a CommitV1 message from the bus
        :returns: True if the message is about a fork, not upstream
        """
        commit = message.body["commit"]
        namespace = commit["namespace"]
        repo = commit["repo"]

        if namespace:
            repo_with_ns = namespace + "/" + repo
        else:
            repo_with_ns = repo

        orig_path = commit["path"]
        path = orig_path.rstrip("/")

        if path.endswith(".git"):
            path = path[:-4]

        path = path.rstrip("/")

        if path.endswith(f"/{repo_with_ns}"):
            return "/fork/" in path[: -len(repo_with_ns)]

        log.warning(
            "Message %s mismatch: path=%r namespace=%r repo=%r",
            message.id,
            orig_path,
            namespace,
            repo,
        )

        # Something is off, rather notify superfluously.
        return False

    def process(self, config: dict[str, Any], message: Message) -> None:
        """Process a given message.

        :arg config: Toddlers configuration
        :arg message: Message to process
        """
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("Processing message:\n%s", json.dumps(message.body, indent=2))

        if not isinstance(message, CommitV1):
            log.debug(
                "Skipping message %s of incompatible type: %s",
                message.id,
                type(message),
            )
            return

        if self.ignore_commit(message):
            log.debug("Ignoring message %r", message)
            return

        # These must be set
        mail_from: Optional[str] = config.get("mail_from")
        mail_to: Optional[str] = config.get("mail_to")
        mail_server: Optional[str] = config.get("mail_server")

        if not mail_from or not mail_to or not mail_server:
            raise ConfigError(
                "Invalid toddler configuration: mail_from, mail_to and mail_server have to be set."
            )

        # These have good default fallbacks
        mail_subject_tmpl: Optional[str] = config.get("mail_subject_tmpl")
        mail_content_tmpl: Optional[str] = config.get("mail_content_tmpl")

        tmpl_params = {
            "message": message,
            "headers": message._headers,
            "body": message.body,
            "commit": message.body["commit"],
        }

        if mail_subject_tmpl:
            subject = mail_subject_tmpl.format(**tmpl_params)
        else:
            subject = message.summary

        if mail_content_tmpl:
            content = mail_content_tmpl.format(**tmpl_params)
        else:
            content = str(message)

        send_email(
            to_addresses=[mail_to],
            from_address=mail_from,
            subject=subject,
            content=content,
            mail_server=mail_server,
        )
