"""
This plugin handles fedora-messaging messages published under the topic
``ci.pipeline.allpackages-pr.complete/running/error`` and flag the
corresponding PR in dist-git with the result of the tests.

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import hashlib
import logging

from ..base import ToddlerBase
from ..utils.requests import make_session


_log = logging.getLogger(__name__)

done_states = {
    "passed": {"api": "success", "human": "passed"},
    "error": {"api": "error", "human": "errored"},
    "failed": {"api": "failure", "human": "failed"},
    "needs_inspection": {"api": "failure", "human": "failed"},
}


class FlagCIPR(ToddlerBase):
    """Listens to messages sent by fedora CI and flags pull-requests on
    src.fedoraproject.org accordingly.
    """

    name = "flag_ci_pr"

    amqp_topics = [
        "org.centos.*.ci.dist-git-pr.test.error",
        "org.centos.*.ci.dist-git-pr.test.complete",
        "org.centos.*.ci.dist-git-pr.test.running",
    ]

    def __init__(self):
        super().__init__()
        self.requests_session = make_session()

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return (
            topic.startswith("org.centos.")
            and "ci.dist-git-pr.test" in topic
            and topic.endswith(("complete", "error", "running"))
        )

    def process(self, config, message):
        """Process a given message."""

        pipeline_state = None
        if "." in message.topic:
            pipeline_state = message.topic.rsplit(".", 1)[-1]

        if pipeline_state not in ["complete", "running", "error"]:
            _log.info("Pipeline state is not 'complete' or 'running' or 'error'.")
            return

        msg = message.body

        if "version" not in msg or not msg["version"].startswith(("0.2.", "1.")):
            _log.info("Unsupported msg version, ignoring")
            return

        if "category" not in msg["test"]:
            _log.info(
                "Invalid message send, no `category` in the `test` section, ignoring - id: %s",
                message.id,
            )
            return
        test_category = msg["test"]["category"]

        if "type" not in msg["test"]:
            _log.info(
                "Invalid message send, no `type` in the `test` section, ignoring - id: %s",
                message.id,
            )
            return
        test_type = msg["test"]["type"]

        pr_flag_title = None
        if test_category == "functional":
            if test_type == "tier0":
                pr_flag_title = "Fedora CI - dist-git test"
            else:
                pr_flag_title = f"Fedora CI - {test_type}"

        if test_category == "static-analysis" and test_type == "build":
            pr_flag_title = "Fedora CI - scratch build"

        # this is the same test as "build.static-analysis"
        if test_category == "validation" and test_type == "scratch-build":
            pr_flag_title = "Fedora CI - scratch build"

        if not pr_flag_title:
            _log.info(
                "Invalid test case, `category` == `%s` and `type` == `%s`, ignoring - id: %s",
                test_category,
                test_type,
                message.id,
            )
            return

        # test complete messages
        if pipeline_state == "complete":
            if "result" not in msg["test"]:
                _log.info(
                    "Invalid message send, no `result` in the `test` section, ignoring - id: %s",
                    message.id,
                )
                return

            state = msg["test"]["result"]
            if state not in done_states:
                _log.info("Build is not in one of the expected status, ignoring")
                return

            status = done_states[state]["api"]
            human_status = done_states[state]["human"]

        # test running messages
        elif pipeline_state == "running":
            status = "pending"
            human_status = "running"

        # test error messages
        elif pipeline_state == "error":
            status = "error"
            human_status = "errored"

        if "id" not in msg["artifact"]:
            _log.info(
                "Invalid message: %s, could not extract the PR id from it"
                % msg["artifact"]
            )
            return

        if "commit_hash" not in msg["artifact"] or not msg["artifact"]["commit_hash"]:
            _log.info("No commit hash specified, bailing: %s" % msg["artifact"])
            return

        pr_id = str(msg["artifact"]["id"])

        commit_hash = msg["artifact"]["commit_hash"]
        commit_hash_text = " for %s" % commit_hash[:8]
        seed = config.get("dist_git_token_seed", config.get("pagure_token_seed")) or ""

        data = {
            "username": pr_flag_title,
            "status": status,
            "comment": "Package tests%s: %s" % (commit_hash_text, human_status),
            "url": msg["run"]["url"],
            "uid": hashlib.md5(
                (pr_id + seed + pr_flag_title).encode("utf-8")
            ).hexdigest(),
        }

        dist_git_url = config.get("dist_git_url") or config["pagure_url"]

        namespace = msg["artifact"]["repository"].split("/")[-2]
        repo = msg["artifact"]["repository"].split("/")[-1]

        target_url = "/".join(["api", "0", namespace, repo, "c", commit_hash, "flag"])
        flag_url = dist_git_url + "/" + target_url
        _log.info("Flagging commit at: %s" % flag_url)

        headers = {
            "Authorization": "token " + config.get("dist_git_token")
            or config["pagure_token"],
            "User-Agent": "loopabull@fedora-infra",
        }

        _log.info("payload: %s" % data)

        req = self.requests_session.request(
            method="POST", url=flag_url, headers=headers, data=data
        )
        _log.info("Request to %s returned: %s" % (dist_git_url, req.status_code))
        if not req.ok:
            _log.debug(req.text)
            return 1
        else:
            _log.info("All clear")
            _log.info(
                "User-URL: %s/" % msg["artifact"]["repository"]
                + "/".join(["pull-request", pr_id])
            )
