"""
This script takes as input the fedmsg messages published under the topic
``buildsys.build.state.change`` and flag the corresponding commit in pagure
if the build was successful.

Authors:    Pierre-Yves Chibon <pingou@pingoured.fr>

"""

import logging

import koji

from ..base import ToddlerBase
from ..utils.requests import make_session


_log = logging.getLogger(__name__)

# see koji.TASK_STATES for all values
done_states = {
    1: "completed",
    3: "failed",
    4: "canceled",
}


def _koji_link(config, msg):
    base_koji_url = config["koji_url"].rstrip("/")

    if not msg["task_id"]:
        return ""

    return f"{base_koji_url}/koji/taskinfo?taskID={msg['task_id']}"


class FlagCommitBuild(ToddlerBase):
    """Listens to messages sent by fedora CI and flags pull-requests on
    src.fedoraproject.org accordingly.
    """

    name = "flag_commit_build"

    amqp_topics = [
        "org.fedoraproject.*.buildsys.build.state.change",
    ]

    def __init__(self):
        super().__init__()
        self.requests_session = make_session()

    def accepts_topic(self, topic):
        """Returns a boolean whether this toddler is interested in messages
        from this specific topic.
        """
        return topic.startswith("org.fedoraproject.") and topic.endswith(
            "buildsys.build.state.change"
        )

    def process(self, config, message):
        """Process a given message."""

        msg = message.body

        # Check if the owner field is even available
        if "owner" not in msg:
            return

        if msg["owner"] == "containerbuild":
            _log.info("Skipping container build")
            return

        instance = msg["instance"]
        if instance != "primary":
            _log.info("Ignoring secondary arch task...")
            return

        state = msg["new"]
        if state not in done_states:
            _log.info("Build is not in a state we care about, skipping")
            return

        base_koji_url = config["koji_url"].rstrip("/")
        koji_url = f"{base_koji_url}/kojihub"
        _log.debug("Talking to koji at: %s" % koji_url)
        koji_client = koji.ClientSession(koji_url)

        build_id = msg["build_id"]
        build = koji_client.getBuild(build_id)
        git_url = (build.get("extra") or {}).get("source", {}).get("original_url")

        if not git_url:
            _log.info(
                "No git url found in the extra information: %s" % build.get("extra")
            )
            return

        if "#" not in git_url:
            _log.info("No # in the git_url: %s" % git_url)
            return

        commit = git_url.rsplit("#", 1)[1]

        if state == 1:
            status = "success"
        elif state == 3:
            status = "failure"
        else:
            status = "canceled"

        version = msg["version"]
        if msg.get("epoch"):
            version = "%s:%s" % (msg["epoch"], version)

        nvr = "%s-%s-%s" % (msg["name"], version, msg["release"])
        data = {
            "username": "Build %s" % (done_states[state]),
            "status": status,
            "comment": "Built as %s" % (nvr),
            "url": _koji_link(config, msg),
        }

        dist_git_url = config.get("dist_git_url") or config["pagure_url"]

        target_url = "/".join(["api", "0", "rpms", msg["name"], "c", commit, "flag"])
        flag_url = dist_git_url + "/" + target_url
        _log.info("Flagging commit at: %s" % flag_url)

        headers = {
            "Authorization": "token " + config.get("dist_git_token")
            or config["pagure_token"],
            "User-Agent": "toddlers@fedora-infra",
        }

        _log.debug("payload: %s" % data)

        req = self.requests_session.request(
            method="POST", url=flag_url, headers=headers, data=data
        )
        _log.info("Request to %s returned: %s" % (dist_git_url, req.status_code))
        if not req.ok:
            _log.debug("  --  FAILED  --  ")
            _log.debug(req.text)
            # raise Exception("Did not successfully flag the build in dist-git")
        else:
            _log.info("All clear")
            _log.info(
                "User-URL: %s" % dist_git_url
                + "/"
                + "/".join(["rpms", msg["name"], "c", commit])
            )
