"""
This toddler synchronizes groups from FAS to pagure.io.

Authors:    Michal Konecny <mkonecny@redhat.com>

"""

import argparse
import json
import logging
import sys
from typing import Dict

from fedora_messaging.api import Message

from toddlers.base import ToddlerBase
from toddlers.exceptions import PagureError
from toddlers.utils import fedora_account, pagure

_log = logging.getLogger(__name__)


class PagureFASGroupsSync(ToddlerBase):
    """
    Synchronize FAS groups with Pagure.

    Consumes group membership change messages and process them
    and consumes trigger to sync all the configured groups.
    """

    name: str = "pagure_fas_groups_sync"

    amqp_topics: list = [
        "org.fedoraproject.*.ipa.group_add_member.v1",
        "org.fedoraproject.*.toddlers.trigger.pagure_fas_groups_sync",
    ]

    # Pagure object
    pagure: pagure.Pagure

    # Group mapping from FAS to pagure
    group_map: Dict[str, str]

    def accepts_topic(self, topic: str) -> bool:
        """
        Return a boolean if this toddler consumes messages from topic.

        :arg topic: Topic to check.

        :returns: True if topic is accepted, False otherwise.
        """
        if topic.startswith("org.fedoraproject."):
            if topic.endswith("ipa.group_add_member.v1"):
                return True
            if topic.endswith("toddlers.trigger.pagure_fas_groups_sync"):
                return True

        return False

    def process(
        self,
        config: dict,
        message: Message,
    ) -> None:
        """Process a given message.

        :arg config: Toddlers configuration
        :arg message: Message to process
        """
        if _log.isEnabledFor(logging.DEBUG):
            _log.debug("Processing message:\n%s", json.dumps(message.body, indent=2))
        topic = message.topic

        self.pagure = pagure.set_pagure(config)
        fedora_account.set_fasjson(config)

        group_map = config.get("group_map", {})

        if topic.endswith("toddlers.trigger.pagure_fas_groups_sync"):
            for group in group_map:
                self.sync_group(group, group_map[group])

        if topic.endswith("ipa.group_add_member.v1"):
            for user in message.user_names:
                if message.group in group_map:
                    try:
                        if self.pagure.user_exists(user):
                            self.pagure.add_member_to_group(
                                user, group_map[message.group]
                            )
                    except PagureError as e:
                        _log.exception(str(e))
                else:
                    _log.info(
                        "User '%s' was added to group '%s', "
                        "but we don't care about this group. "
                        "Skipping.",
                        user,
                        message.group,
                    )

    def sync_group(self, fas_group: str, pagure_group: str):
        """
        Synchronize FAS group with Pagure group.

        Compares FAS group and Pagure group and modifies the Pagure
        group to correspond with the specified FAS group.

        :arg fas_group: FAS group to check
        :arg pagure_group: Pagure group to sync
        """
        _log.info(
            "Syncing FAS group '%s' with pagure group '%s'", fas_group, pagure_group
        )
        group_members_fas = fedora_account.get_group_member(fas_group)
        _log.debug("FAS group members: [%s]", group_members_fas)
        group_members_pagure = self.pagure.get_group_members(pagure_group)
        _log.debug("Pagure group members: [%s]", group_members_pagure)

        if not group_members_fas or not group_members_pagure:
            _log.warning(
                "FAS group or Pagure group is empty. This shouldn't happen. Skipping sync."
            )
            return
        add_members = [
            user for user in group_members_fas if user not in group_members_pagure
        ]
        remove_members = [
            user for user in group_members_pagure if user not in group_members_fas
        ]

        for user in add_members:
            try:
                if self.pagure.user_exists(user):
                    self.pagure.add_member_to_group(user, pagure_group)
            except PagureError as e:
                _log.exception(str(e))

        for user in remove_members:
            try:
                if self.pagure.user_exists(user):
                    self.pagure.remove_member_from_group(user, pagure_group)
            except PagureError as e:
                _log.exception(str(e))

        _log.info("Sync complete")


def main(args):
    """Run toddler class without trigger message."""
    parser = argparse.ArgumentParser(
        description="Sync the group membership from FAS to pagure."
    )
    parser.add_argument(
        "--fas-url",
        default="https://fasjson.fedoraproject.org",
        help="URL to the FAS instance to use.",
    )
    parser.add_argument(
        "--pagure-url",
        default="https://pagure.io",
        help="URL to the Pagure instance to use.",
    )
    parser.add_argument(
        "--api-key",
        help="Pagure API key to use.",
        required=True,
    )
    parser.add_argument(
        "--group",
        help="Sync a specific group from FAS to pagure.",
        required=True,
    )
    parser.add_argument(
        "--target-group",
        help="Name of the group on the pagure side that the group specified "
        "in --group should be synced to.",
        required=True,
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Print the debugging output",
    )
    args = parser.parse_args(args)
    _log.addHandler(logging.StreamHandler(sys.stdout))
    if args.debug:
        _log.setLevel(logging.DEBUG)

    pagure_fas_sync_toddler = PagureFASGroupsSync()
    pagure_fas_sync_toddler.pagure = pagure.set_pagure(
        {"pagure_url": args.pagure_url, "pagure_api_key": args.api_key}
    )
    fedora_account.set_fasjson({"fas_url": args.fas_url})
    pagure_fas_sync_toddler.sync_group(args.group, args.target_group)


if __name__ == "__main__":  # pragma: no cover
    try:
        main(sys.argv[1:])
    except KeyboardInterrupt:
        pass
