"""
This module is a wrapper for Anitya. It uses Anitya API to communicate and configure
release monitoring instance.
To work with it, you need to set it up by calling `set_anitya`.

Examples:
    from utils import anitya

    anitya_config = {
        "anitya_endpoint": "https://release-monitoring.org/",
        "anitya_access_token": "secret TOKEN",
    }

    anitya_obj = anitya.set_anitya(config)
    anitya_obj.create_project_in_anitya("<name>", "<homepage>", "<backend>")
"""

import logging
from typing import Optional

from toddlers.utils import requests

log = logging.getLogger(__name__)


def set_anitya(config):
    """
    Set the connection to the Anitya API.

    Params:
        config: Configuration dictionary.
    """
    return Anitya(config)


class Anitya(object):
    """
    Object that works with Anitya.
    """

    # URL to Anitya
    _anitya_endpoint: str = ""
    # API TOKEN to Anitya
    _anitya_access_token: Optional[str] = None
    # Request Session object used for communication
    _requests_session: requests.requests.Session

    def __init__(self, config):
        """
        Initialize the Anitya class.

        Params:
            config (dict): A configuration with anitya_endpoint and anitya_access_token keys.

        Raises:
            ValueError: If no pagure_api_key is provided.
        """
        self._anitya_endpoint = config.get("anitya_endpoint", "").removesuffix("/")
        if not self._anitya_endpoint:
            raise ValueError("No anitya endpoint found in config file")

        self._anitya_token = config.get("anitya_access_token", "")
        if not self._anitya_token:
            raise ValueError("No anitya access token found in config file")

        self._requests_session = requests.make_session(timeout=300)

    def does_project_exists_in_anitya(self, project_name: str) -> Optional[str]:
        """
        Check if project exists in Anitya.

        Params:
            project_name (str): The name of the project.

        Returns:
            Optional[str]: project URL if it exists in Anitya, otherwise None.
        """
        projects_params = {
            "name": project_name,
        }
        endpoint = self._anitya_endpoint + "/api/v2/projects/"
        projects_response = self._requests_session.get(endpoint, params=projects_params)
        if projects_response.status_code != 200:
            log.debug("Project '{0}' not found in Anitya.".format(project_name))
            return None

        response_json = projects_response.json()
        item_count = response_json["total_items"]
        if item_count == 0:
            log.debug("Project '{0}' not found in Anitya.".format(project_name))
            return None

        try:
            project_id = response_json["items"][0]["id"]
            project_url = "{0}/project/{1}".format(self._anitya_endpoint, project_id)
            return project_url
        except (KeyError, IndexError):
            return None

    def does_package_exists_in_anitya(
        self, package_name: str, distribution: str, project_name: str
    ) -> bool:
        """
        Check if package exists in Anitya.

        Params:
            package_name (str): The name of the package.
            distribution (str): The name of the distribution.
            project_name (str): The name of the project.

        Returns:
            False if package don't exist
            False if package exist but his project is different from provided project name
            True if package exist and his project is correct
        """
        endpoint = self._anitya_endpoint + "/api/v2/packages/"
        packages_params = {
            "name": package_name,
            "distribution": distribution,
        }
        packages_response = self._requests_session.get(endpoint, params=packages_params)
        if packages_response.status_code != 200:
            log.info("Package '{0}' not found in Anitya.".format(package_name))
            return False  # Not able to find package
        response_json = packages_response.json()
        item_count = response_json["total_items"]
        if item_count < 1:
            log.info("Package '{0}' not found in Anitya.".format(package_name))
            return False
        package_json = response_json["items"][0]
        package_project_name = package_json["project"]
        if package_project_name != project_name:
            return False  # Expected and actual project name are different
        else:
            return True  # Expected and actual project name are the same

    def create_project_in_anitya(
        self,
        name: str,
        homepage: str,
        backend: str,
    ) -> Optional[str]:
        """
        Create a new project in Anitya.

        Params:
            name (str): The name of the project.
            homepage (str): The homepage of the project.
            backend (str): The name of the backend.

        Returns:
            The project URL if successful, otherwise None.
        """
        headers = {"Authorization": "token " + self._anitya_token}
        endpoint = self._anitya_endpoint + "/api/v2/projects/"
        payload = {
            "name": name,
            "homepage": homepage,
            "backend": backend,
        }
        log.info("Creating project '{0}' in Anitya.".format(name))
        response = self._requests_session.post(
            url=endpoint, data=payload, headers=headers
        )
        if response.status_code == 201:
            project_json = response.json()
            project_id = project_json["id"]
            project_url = "{0}/project/{1}".format(self._anitya_endpoint, project_id)
            return project_url
        else:
            return None

    def create_package_in_anitya(
        self,
        package_name: str,
        project_name: str,
        distribution: str,
        project_ecosystem: str,
    ) -> Optional[str]:
        """
        Create a new package in Anitya.

        Params:
            package_name (str): The name of the package.
            project_name (str): The name of the project.
            distribution (str): The name of the distribution.
            project_ecosystem (str): The name of the ecosystem.

        Returns:
            Return message if status code is known, otherwise None.
        """
        headers = {"Authorization": "token " + self._anitya_token}
        endpoint = self._anitya_endpoint + "/api/v2/packages/"
        payload = {
            "package_name": package_name,
            "project_name": project_name,
            "distribution": distribution,
            "project_ecosystem": project_ecosystem,
        }
        log.info("Creating package '{0}' in Anitya.".format(package_name))
        response = self._requests_session.post(
            url=endpoint, data=payload, headers=headers
        )
        if response.status_code == 400:
            return "Bad Request, some necessary arguments were not provided."
        elif response.status_code == 401:
            return "Unauthorized, access token is incorrect."
        elif response.status_code == 409:
            return "Conflict, package already exists."
        elif response.status_code == 201:
            return "Success"
        return None
