"""
This module is a wrapper for bodhi. It uses requests to call configured bodhi endpoints.
To work with it you need to set it up first by calling `set_bodhi`.

Examples:
  from toddlers.utils import bodhi

  bodhi_config = {
    "bodhi_url": "https://bodhi.fedoraproject.org",
  }

  bodhi_instance = bodhi.set_bodhi(bodhi_config)

  bodhi_instance.get_active_branches()
"""

import logging
from typing import Any

from toddlers.exceptions import BodhiError
from toddlers.utils import requests

log = logging.getLogger(__name__)


def set_bodhi(conf: dict) -> Any:
    """
    Set the connection to bodhi.

    Params:
      conf: This needs to have key "bodhi_url"
    """
    return Bodhi(conf)


def _handle_response_failures(response, action, api_url) -> BodhiError:
    """Factors out handling of non-200 responses from Bodhi. Returns
    an error for the calling method to raise (this makes mypy happy).
    """
    log.error(
        "Error when trying to %s. Got status_code '%s'.",
        action,
        response.status_code,
    )

    response_json = None
    if response.headers.get("content-type") == "application/json":
        response_json = response.json()
        log.error("Received response: %s", response.json())

    return BodhiError(
        (
            "Couldn't {0}.\n\n"
            "Request to '{1}':\n\n"
            "Response:\n"
            "{2}\n\n"
            "Status code: {3}"
        ).format(
            action,
            api_url,
            response_json,
            response.status_code,
        )
    )


class Bodhi:
    """Object that is a wrapper above bodhi endpoints."""

    # URL for bodhi
    _bodhi_url: str = ""
    # Requests Session object used for communication
    _requests_session: requests.requests.Session

    def __init__(self, config: dict) -> None:
        """
        Initialize bodhi instance.

        Params:
          config: This needs to have keys "bodhi_url"

        Raises:
          `ValueError` if configuration value is missing
        """
        self._bodhi_url = config.get("bodhi_url", "")
        if not self._bodhi_url:
            raise ValueError("No bodhi_url found in the configuration file")

        self._requests_session = requests.make_session(timeout=60)

    def get_active_branches(self) -> list:
        """
        Get active branches from bodhi.

        Returns:
          Dictionary containing active branches.

        Raises:
          toddlers.utils.exceptions.BodhiError when the issue couldn't be retrieved.
        """
        api_url = "{0}/releases/?exclude_archived=True".format(self._bodhi_url)
        log.debug("Retrieving active branches from bodhi.")
        response = self._requests_session.get(api_url)

        result = []

        if response.status_code == 200:
            response_json = response.json()
            try:
                if response_json["releases"]:
                    for branch in response_json["releases"]:
                        result.append(branch["branch"])
                    return result
            except KeyError:
                raise BodhiError(
                    (
                        "Expected key not found in bodhi response. "
                        "Expected data: {{'releases': [{{'branch': <branch_name>}}]}}\n\n"
                        "Request to '{0}':\n\n"
                        "Response:\n"
                        "{1}\n\n"
                        "Status code: {2}"
                    ).format(
                        api_url,
                        response_json,
                        response.status_code,
                    )
                )

        raise _handle_response_failures(response, "retrieve active branches", api_url)

    def get_branch_info(self, branch) -> dict:
        """
        Get info about a particular branch from bodhi.
        Params:
            branch: string

        Returns:
          Dict containing branch info.

        Raises:
          toddlers.utils.exceptions.BodhiError when the issue couldn't be retrieved.
        """
        api_url = "{0}/releases/{1}".format(self._bodhi_url, branch.upper())
        log.debug("Retrieving branch info from bodhi.")
        response = self._requests_session.get(api_url)

        if response.status_code == 200:
            # the response is always a dict, but mypy doesn't know that
            return dict(response.json())

        raise _handle_response_failures(
            response, "get branch info for {0}".format(branch), api_url
        )

    def is_branch_frozen(self, branch) -> bool:
        """Check whether the given branch is currently frozen.
        Params:
            branch: string

        Returns:
            bool
        """
        try:
            branch_info = self.get_branch_info(branch)
        except BodhiError:
            return False
        # why bool? see https://stackoverflow.com/questions/75864807
        return bool(branch_info.get("state", "unknown") == "frozen")

    def is_branch_stable(self, branch) -> bool:
        """Check whether the given branch is already stable, based on its 'state'.
        Params:
            branch: string

        Returns:
            bool
        """
        try:
            branch_info = self.get_branch_info(branch)
        except BodhiError:
            return True
        return branch_info.get("state", "unknown") in ("current", "archived", "unknown")
