"""
Wrapper for git operations using GitPython library.

Author: mkonecny@redhat.com
"""

import os
from typing import Optional, Union

import git


def clone_repo(remote: str, destination: str) -> "GitRepo":
    """
    Clone the remote repository.

    Params:
      remote: URL of remote repository
      destination: Path to clone the repo to

    Returns:
      Object representing git repository.
    """
    repo = git.Repo.clone_from(remote, destination)

    return GitRepo(repo)


class GitRepo:
    """
    Class representing git repository.

    Arguments:
      repo (`git.Repo`): Inner representation of git repository
    """

    # repo: Optional[git.Repo] = None

    def __init__(self, repo: git.Repo) -> None:
        """
        Initialize GitRepo object.

        Params:
          repo: Git repo object
        """
        self.repo = repo

    def first_commit(self, branch: str) -> Optional[str]:
        """
        Returns the first commit on the repository on specified branch.

        Params:
          branch: Branch to look for the first commit on

        Returns:
          Hash of commit.
        """
        # This will return only commit hashes for commits
        commits = self.repo.git.log("--reverse", "--pretty=format:%H", branch)  # type: ignore

        result = None

        if commits:
            result = commits.split("\n")[0]

        return result

    def does_branch_exist(self, branch_name: str) -> bool:
        """
        Checks if the branch exists in the repository

        :param branch_name: The name of the branch to check

        :return: True if the branch exists, False otherwise
        """
        remote = self.repo.remotes.origin
        for ref in remote.refs:
            if remote.name + "/" + branch_name == ref.name:
                return True
        return False

    def does_branch_contains_file(self, branch_name: str, file_path: str) -> bool:
        """
        Checks if the existing branch contain specific file path.

        :param branch_name: The branch name to check
        :param file_path: The path of the file to check

        :return: True if the file exists, False otherwise
        """
        branch_name = "origin/" + branch_name
        try:
            commit = self.repo.commit(branch_name)
            for blob in commit.tree:
                if blob.name == file_path:
                    return True
            return False
        except ValueError:  # Branch doesn't exist
            return False

    def get_last_commit_message(
        self, branch: str = "rawhide"
    ) -> Union[str, bytes, None]:
        """
        Returns the message of the last commit on specific branch.

        :param branch: Branch to look for the last commit

        Returns:
            Commit message.
        """
        try:
            if branch != "rawhide":
                if not branch.startswith("origin/"):
                    branch = "origin/" + branch
            commit = self.repo.commit(branch)
            return commit.message
        except Exception:  # Raised when branch name is not correct
            return None

    def get_last_commit_date(self, branch: str = "rawhide") -> Optional[int]:
        """
        Returns the timestamp of the last commit on specific branch.

        :param branch: Branch to look for the last commit

        Returns:
            Timestamp of last commit.
        """
        try:
            if branch != "rawhide":
                if not branch.startswith("origin/"):
                    branch = "origin/" + branch
            commit = self.repo.commit(branch)
            return commit.committed_date
        except Exception:  # Raised when branch name is not correct
            return None

    def revert_last_commit(
        self, message: str, user: str, token: str, branch: str = "rawhide"
    ) -> None:
        """
        Revert last commit with message on requested branch.

        Params:
            message: The message we would like to revert the last commit with.
            branch: The branch to revert the last commit on
        """
        try:
            # checkout to the correct branch
            self.repo.git.checkout(branch)

            # reverting last commit and changing the commit message
            self.repo.git.revert("HEAD", no_edit=True)
            self.repo.git.commit("--amend", "-m", message)

            origin = self.repo.remote("origin")
            git_cmd = self.repo.git

            push_url = origin.url.replace(
                "https://", "https://{0}:{1}@".format(user, token)
            )
            git_cmd.push("-u", push_url, branch)
        except Exception as error:
            print(error, "\nSomething happened during reverting the last commit")

    def retire_branch(
        self, message: str, user: str, token: str, branch: str = "rawhide"
    ):
        """
        Retire branch on git repository. This will create dead.package on
        branch and pushes the changes.

        Params:
            message: Content of dead.package.
            user: User pushing the changes
            token: ACL token to use for authentication
            branch: The branch to retire
        """
        # checkout to the correct branch
        self.repo.git.checkout(branch)

        # remove all files first
        for entry in self.repo.commit().tree:
            self.repo.index.remove(entry.name, working_tree=True)

        file_name = os.path.join(str(self.repo.working_tree_dir), "dead.package")
        with open(file_name, "w") as file:
            file.write(message)
        self.repo.index.add(["dead.package"])

        self.repo.index.commit("Retire branch")

        origin = self.repo.remote("origin")
        git_cmd = self.repo.git

        push_url = origin.url.replace(
            "https://", "https://{0}:{1}@".format(user, token)
        )
        git_cmd.push("-u", push_url, branch)
