from email.message import EmailMessage
import logging
import smtplib


_log = logging.getLogger(__name__)


def notify_packager(mail_server, admin_email, username, email):
    message = f"""Hello {username},

We have identified you[1] as either a Fedora packager or someone who has asked to
be included in the CC list of tickets created for one or more component on
bugzilla. Fedora packagers are granted special permissions on the Fedora bugs in
bugzilla.
However, to enable these functionalities (granting you these permissions or
including you to the CC list of your packages of interest), we need to have your
bugzilla email address stored in the Fedora Account System[2].
At the moment you have:

{email}

which bugzilla is telling us is not an account in bugzilla.  If you could
please set up an account in bugzilla with this address or change your email
address (either Red Hat Bugzilla email or Fedora Account email) on your Fedora
Account to match an existing bugzilla account this would let us go forward.

Note #1: this message is being generated by an automated script.  You'll
continue getting this message until the problem is resolved.  Sorry for the
inconvenience.

Thank you,
The Fedora Account System
{admin_email}


[1] the source of this information is the following JSON file:
    https://src.fedoraproject.org/extras/pagure_bz.json
    We are happy to tell you exactly which packages are linked to your account
    if you wish.
[2] https://accounts.fedoraproject.org/
"""

    send_email(
        to_addresses=[email],
        from_address=admin_email,
        subject="Fedora Account System and Bugzilla Mismatch",
        content=message,
        mail_server=mail_server,
    )


def notify_admins(mail_server, admin_email, recipients, users):
    people = []
    for person in users:
        if person.person.status == "Active":
            people.append(
                "  %(user)s  --  %(name)s  --  %(email)s"
                % {
                    "name": person.person.human_name,
                    "email": person.email,
                    "user": person.person.username,
                }
            )
    if people:
        people = "\n".join(people)
        message = (
            """
The following people are in the packager group but do not have email addresses
that are valid in bugzilla:
%s

"""
            % people
        )

    send_email(
        to_addresses=recipients,
        from_address=admin_email,
        subject="Fedora Account System and Bugzilla Mismatch",
        content=message,
        mail_server=mail_server,
    )


def notify_admins_on_packagers_without_bugzilla_accounts(
    to_addresses, from_address, mail_server, logs_text
):
    message = f"""Dear Admin,

The packagers_without_bugzilla toddler just ran and noticed some packagers without
valid bugzilla account. The list of them is here:
- {logs_text}

Each person on this list has been notified and, hopefully will fix the situation.
Failing to do so, this notification may be used a record for a potential
non-responsive packager procedure.

Have a wonderful day and see you (maybe?) at the next run!

"""
    send_email(
        to_addresses=to_addresses,
        from_address=from_address,
        subject="Toddlers found some packagers without bugzilla account",
        content=message,
        mail_server=mail_server,
    )


def notify_admins_distgit_sync_error(mail_server, admin_email, recipients, errors):
    """Send e-mail to admins about errors encountered during synchronization of bugzilla
    and distgit.

    Params:
        mail_server (str): Mail server to sent e-mail from
        admin_email (str): Admin e-mail to use as sender
        recipients (list): List containing e-mails of recipients
        errors (list): List of errors to report
    """
    errors = "\n".join(errors)
    message = f"""Greetings, our mighty admins,

While updating bugzilla with information from the Package Database some errors were encountered.
Don't panic!
Please, be so kind and see if you can do something with them. Here is the list of the errors:

{errors}
"""

    send_email(
        to_addresses=recipients,
        from_address=admin_email,
        subject="Errors while syncing bugzilla with the PackageDB",
        content=message,
        mail_server=mail_server,
    )


def notify_admins_on_retired_packages_cleanup(
    to_addresses, from_address, mail_server, logs_text
):
    message = f"""Dear Admin,

The clean_retired_packages toddler just ran and adjusted some retired packages
for recording purposes, here is what it did:
- {logs_text}

In case someone asks, you can come back to this email but there isn't really
anything else to do with it.

Have a wonderful day and see you (maybe?) at the next run!

"""
    send_email(
        to_addresses=to_addresses,
        from_address=from_address,
        subject="Toddlers cleaned up some retired packages",
        content=message,
        mail_server=mail_server,
    )


def notify_admins_on_invalid_entry_in_email_override(
    to_addresses, from_address, mail_server, url, logs_text
):
    message = f"""Dear Admin,

In the email_overrides.toml file located at:
{url}

We have identified the following issues:
- {logs_text}

Overrides having no FAS account can be simply removed.

Overrides having no bugzilla account needs to be notified to the person.


Have a wonderful day and see you (maybe?) at the next run!

"""
    send_email(
        to_addresses=to_addresses,
        from_address=from_address,
        subject="Invalid entries in email_overrides.toml",
        content=message,
        mail_server=mail_server,
    )


def send_email(
    to_addresses, from_address, subject, content, mail_server, cc_address=None
):
    """Actually sends the email to the list of addresses specified from the
    address given with the subject and content and via the given email server.
    """

    _log.info("Sending email to %s about: %s", to_addresses, subject)
    msg = EmailMessage()
    msg.add_header("To", ", ".join(to_addresses))
    msg.add_header("From", from_address)

    if cc_address is not None:
        msg.add_header("Cc", ",".join(cc_address))
        to_addresses += cc_address

    msg.add_header("Subject", subject)
    msg.set_content(content)
    smtp = smtplib.SMTP(mail_server)
    smtp.send_message(msg, from_address, to_addresses)
    smtp.quit()
    _log.info("email sent")
